import numpy as np
from json import loads
from typing import Union

from ..obj.units import make_unit
from ..utils.manage_strand_reactor_files import read_txt

from .strand_motifs_trajectory import _extract_upper_and_lower_strands_as_continuous_strings_from_structure_of_complex
from .strand_motifs_trajectory import _transform_motif_string_to_index_tuple

from ..obj.times_vector import TimesVector
from ..obj.motif_production_vector import MotifProductionVector
from ..obj.motif_production_vector import _create_empty_motif_production_dict
from ..obj.motif_production_trajectory import MotifProductionTrajectory

from ..domains.motif_production_space import make_motif_production_dct, _production_channel_id
from ..domains.motif_space import _return_motif_categories, _motif_categories

def strand_motifs_productions_trajectory(filepaths : list,
        alphabet : list,
        motiflength : int,
        maximum_ligation_window_length : int
    ) -> MotifProductionTrajectory:
    time_unit = make_unit('t_0')
    motif_production_vectors = []
    total_physical_times = []
    for filepath in filepaths:
        step, current_physical_times, ligstr = _step_simulation_and_ligation_strands_array(filepath)
        total_physical_times = total_physical_times + list(current_physical_times)
        for time_index in range(len(current_physical_times)):
            motif_production_vectors = motif_production_vectors + [_extract_motif_production_vector(ligstr[time_index],
                    motiflength,
                    alphabet,
                    maximum_ligation_window_length)]
    total_physical_times = TimesVector(np.asarray(total_physical_times), time_unit)
    return MotifProductionTrajectory(motif_production_vectors,total_physical_times)

def _calculate_ligation_window_length_and_spot_and_motifs(ending_strand : str,
        leaving_strand : str,
        template_strand_1st_part : str,
        template_strand_2nd_part :str,
        motiflength : int,
        maximum_ligation_window_length : int
        ) -> Union[int,int,str,str,str,str,str]:
    """
    Returns:
    --------
    ligation_window_length : int,
    ligation_spot : int,
    product_motif : str,
    ending_motif : str,
    leaving_motif : str
    template_motif_1st_part : str,
    template_motif_2nd_part : str
    """
    left_ligation_window_length = min(
            len(ending_strand),
            len(template_strand_2nd_part)
            ) + 1
    right_ligation_window_length = min(
            len(leaving_strand),
            len(template_strand_1st_part)
            ) + 1
    ligation_window_length = left_ligation_window_length + right_ligation_window_length
    if ligation_window_length > maximum_ligation_window_length:
        if (left_ligation_window_length
                > (maximum_ligation_window_length -
                    maximum_ligation_window_length//2)):
            if right_ligation_window_length > maximum_ligation_window_length//2:
                left_ligation_window_length = (maximum_ligation_window_length - maximum_ligation_window_length//2)
                right_ligation_window_length = maximum_ligation_window_length//2
            else:
                left_ligation_window_length = (maximum_ligation_window_length
                        - right_ligation_window_length)
        else:
            right_ligation_window_length = maximum_ligation_window_length - left_ligation_window_length
        ligation_window_length = left_ligation_window_length + right_ligation_window_length
    ligation_spot = left_ligation_window_length-1
    ending_motif = ending_strand[-(motiflength-1):]
    leaving_motif = leaving_strand[:(motiflength-1)]
    product_motif = ending_motif[-left_ligation_window_length:] + leaving_motif[:right_ligation_window_length]
    template_motif_1st_part = template_strand_1st_part[-right_ligation_window_length:]
    template_motif_2nd_part = template_strand_2nd_part[:left_ligation_window_length]
    return ligation_window_length, ligation_spot, product_motif, ending_motif, leaving_motif, template_motif_1st_part, template_motif_2nd_part

def _extract_motif_production_vector(ligstr,
        motiflength : int,
        alphabet : list,
        maximum_ligation_window_length : int
        ) -> MotifProductionVector:
    motif_production_dct = _create_empty_motif_production_dict(motiflength,
            alphabet,
            maximum_ligation_window_length)
    for reaction_index in range(len(ligstr)):
        (central_produced_motif,
                ending_motif,
                leaving_motif,
                template_motif,
                production_channel_id
                ) = _motif_production_reactants(
                        *ligstr[reaction_index],
                        motiflength,
                        maximum_ligation_window_length
                        )
        production_index = _transform_motif_string_to_index_tuple(central_produced_motif+template_motif,
                alphabet)
        try:
            motif_production_dct[production_channel_id][production_index] += 1
        except KeyError:
            print(motiflength)
            print(motif_production_dct.keys())
            print(production_index)
            print(central_produced_motif)
            print(template_motif)
            print(production_channel_id)
            raise KeyError
    motif_production_vector = MotifProductionVector(motiflength,alphabet, '1', maximum_ligation_window_length)
    return motif_production_vector(motif_production_dct)

def _load_complex(cc):
    cc = '['+str(cc)+']'
    return cc.replace('”','"').replace('−','-').replace("None","0")

def _read_key(keys : str) -> Union[tuple,str]:
    '''
    returns a tuple of the first ligation statistics stated in keys
    according to the formula below
    and the remaining string

    key: ll-lr|LS-LR|LT|nl-nr|sl-sr|kl-kr|lt|es-ls:n,
    value: #

    ll = length of left segment
    lr = lenght of right segment
    LL = length of left strand (note: segment != strand)
    LR = lnegth of right strand (note: segment != strand)
    LT = length of template strand (note: segment != strand)
    
    nl = number of mismatches of left segment
    nr = number of mismatches of right segment
    sl = number of mismatches left of ligation site
    sr = number of mismatches right of ligation site
    kl = dehyb rate of left segment
    kr = dehyb rate of right segment

    lt = ligation triplex
    es = ending strand
    ls = leaving strand
    n = occupation number?

    PARAMETERS:
    -----------
    keys : string
        string in the format ll-lr|LS-LR|LT|nl-nr|sl-sr|kl-kr|lt|es-ls:n,
        and eventually further ligations afterwards

    RETURN:
    -------
    rtrn : tuple
        in the format (ll,lr,LS,LR,LT,nl,nr,sl,sr,kl,kr,lt,es,ls,n)
    keys : string
        remaining string, i.e. the original string after lstripping the rtrn
        tuple
    '''
    separators = ['-','|','-','|','|','-','|','-','|','-','|[',']|','-',':',',']
    types = [int,int,int,int,int,
            int,int,int,int,np.float64,np.float64,
            _load_complex,str,str,int]
    rtrn = ()
    for ii in range(len(separators)):
        current, _, keys = keys.partition(separators[ii])
        rtrn += (types[ii](current),)
    return rtrn, keys

def _ligation_statistics(fname : str = None,
        skiprows : int = 2) -> Union[np.ndarray,np.ndarray,list]:
    '''
    reads the "ligation_statistics.txt" output file and returns the output as
    lists
    
    PARAMETERS:
    -----------
    fname : string (optional)
        path to the ligation_statistics file
        default : './data/ligation_statistics.txt'
    skiprows : int, optional
        Skip the first `skiprows` lines, including comments
        when reading the file;
        default : 2

    RETURN:
    -------
    step : np.array
        the simulation steps
    total_physical_time : array
        corresponding physical times
    ligation_statistics : list of tuples
        ligation_statistics at corresponding simulation step in the format
        (ll,lr,LS,LR,LT,nl,nr,sl,sr,kl,kr,lt,es,ls,n)
    '''
    replace=[['step = ',''],[' simulation_time = ',';'],['\n{',';{'],[' ',''],[';',' ']]
    x = read_txt(fname, replace=replace)
    x = np.array(x.split()).reshape(-1,3).transpose()
    step = np.array(x[0], dtype = float)
    total_physical_time = np.array(x[1], dtype = float)
    complexes=x[2]
    ligation_statistics = [[]]*len(step)
    for ii in range(len(step)):
        keys = complexes[ii].lstrip('{').rstrip('}')
        ligation_statistics[ii] = []
        while keys != '':
            lg, keys = _read_key(keys)
            ligation_statistics[ii].append(lg)
    return step, total_physical_time, ligation_statistics

def _step_simulation_and_ligation_strands_array(fname : str =None,
        skiprows : int = 2) -> Union[list,list,list]:
    """
    Reads the ligation file and returns its content.

    PARAMETERS:
    -----------
    fname : string
        The filename (and path) of the ligation_statistics file
        default: None
    skiprows : int
        Integer of how many lines are skipped in the ligation_statistics file
        because they indicate the columns
        default : 2

    RETURNS:
    --------
    step : integer list
        Computation steps at which ligations have been measured
    total_physical_time : list
        The physical time at those steps
    ligation_strands : list
        List of tuples of the format
        (ending_strand, leaving_strand, leaving_template, ending_template)
        Like the output of _splitted_template_strand
    """
    step, total_physical_time, ligation_statistics = _ligation_statistics(fname=fname, skiprows=2)
    rtrn = [[]]*len(step)
    for ii in range(len(step)):
        rtrn[ii] = []
        if len(ligation_statistics[ii])!=0:
            for ls in ligation_statistics[ii]:
                binary_complex = ls[-4]
                ending_segment = ls[-3]
                leaving_segment = ls[-2]
                rtrn[ii].append(tuple(
                _strip_zeros(sequence) for sequence in _splitted_template_strand(binary_complex,
                        ending_segment,
                        leaving_segment)
                ))
    return step, total_physical_time, rtrn

def _strip_zeros(sequence : str):
    return sequence.lstrip('0').rstrip('0')

def _ligation_spot(binary_complex, ending_segment, leaving_segment):
    '''
    returns the index of the ligation spot in terms of segments
    '''
    try:
        binary_complex = loads(binary_complex)
    except:
        print("binary_complex cannot be interpreted by json.loads. I assume, it already is a list.")
    # check indices of upper strand
    index_ending_segment = np.where(np.array(binary_complex)==('5'+ending_segment+'3'))
    index_leaving_segment = np.where(np.array(binary_complex)==('5'+leaving_segment+'3'))
    ligation_spot = []
    if len(index_ending_segment[0])>0 and len(index_leaving_segment[0])>0:
        # check that the indices fulfill i2-i1=2
        # and that in between there is a ligation spot
        for ii in index_ending_segment[0][index_ending_segment[1]==0]+1:
            if ii in index_leaving_segment[0][index_leaving_segment[1]==0]-1:
                if binary_complex[ii] == ["|","-"]:
                    ligation_spot.append([ii,0])
    # check indices of lower strand
    index_ending_segment = np.where(np.array(binary_complex)==('3'+ending_segment[::-1]+'5'))
    index_leaving_segment = np.where(np.array(binary_complex)==('3'+leaving_segment[::-1]+'5'))
    if len(index_ending_segment[0])>0 and len(index_leaving_segment[0])>0:
        # check that the indices fulfill i2-i1=2
        # (note that in the lower strand we go from right to left)
        # and that in between there is a ligation spot
        for ii in index_ending_segment[0][index_ending_segment[1]==1]-1:
            if ii in index_leaving_segment[0][index_leaving_segment[1]==1]+1:
                if binary_complex[ii] ==["-","|"]:
                    ligation_spot.append([ii,1])
    if len(ligation_spot) > 1:
        print("Warning: more than one possible ligation spot.")
    elif len(ligation_spot) == 0:
        print("Warning: Did not find any one possible ligation spot.")
        print(binary_complex,"\n", ending_segment,"\n", leaving_segment)
    return np.array(ligation_spot).transpose()

def _splitted_template_strand(binary_complex : str,
        ending_segment : str,
        leaving_segment : str) -> Union[str,str,str,str]:
    '''
    gives the leaving strand, the ending strand and the two parts of the
    template in clockwise direction, i.e.
    ending_strand | leaving_strand
    etalpmet_gnidne - etalpmet_gnivael
    resp. if ligation_happens_in_lower_strand
    leaving_template - ending_template
    dnarts_gniveal | dnarts_gnidne

    PARAMETERS:
    -----------
    binary_complex : string
    ending_segment : string
    leaving_segment : string

    RETURNS:
    --------
    ending_strand : string
        left strand from 5' to 3'
    leaving_strand : string
        right strand from 5' to 3'
    leaving_template : string
        left template strand from 5' to 3'
    ending_template : string
        right template strand from 5' to 3'
    '''
    ligation_spot = _ligation_spot(binary_complex, ending_segment, leaving_segment)
    if ligation_spot[1][0] != 1 and ligation_spot[1][0]!=0:
        raise NotImplementedError("The ligation spot seems to be neither in the lower nor the upper strand")
    if len(ligation_spot[0])>1:
        print("I'll consider the first ligation spot")
    ligation_happens_in_lower_strand = bool(ligation_spot[1][0])
    ligation_spot = ligation_spot[0][0]
    try:
        binary_complex = loads(binary_complex)
    except:
        pass
    try:
        binary_complex = np.array(binary_complex)
    except:
        pass
    left_complexpart = binary_complex[:ligation_spot]
    right_complexpart = binary_complex[ligation_spot+1:]
    if ligation_happens_in_lower_strand:
        leaving_template , leaving_strand = _extract_upper_and_lower_strands_as_continuous_strings_from_structure_of_complex(left_complexpart)
        ending_template, ending_strand = _extract_upper_and_lower_strands_as_continuous_strings_from_structure_of_complex(right_complexpart)
    else:
        ending_strand, ending_template = _extract_upper_and_lower_strands_as_continuous_strings_from_structure_of_complex(left_complexpart)
        leaving_strand, leaving_template = _extract_upper_and_lower_strands_as_continuous_strings_from_structure_of_complex(right_complexpart)
    ending_strand = _strip_zeros(ending_strand).split('0')[-1]
    leaving_strand = _strip_zeros(leaving_strand).split('0')[0]
    leaving_template = _strip_zeros(leaving_template).split('0')[-1]
    ending_template = _strip_zeros(ending_template).split('0')[0]
    return ending_strand, leaving_strand, leaving_template, ending_template

def _determine_production_channel_id(
        ending_motif : str,
        leaving_motif : str,
        template_motif_1st_part : str,
        template_motif_2nd_part : str,
        ligation_window_length : int,
        ligation_spot : int,
        motiflength:int,
        )->str:
    motif_categories = [motif_category.format(ligation_window_length-2) for motif_category in _motif_categories()]
    if len(ending_motif)<ligation_spot+1:
        product_begins = True
    else:
        product_begins = False
    if len(leaving_motif)<ligation_window_length-(ligation_spot+1):
        product_ends = True
    else:
        product_ends = False
    if len(template_motif_2nd_part)<ligation_spot+1:
        template_ends = True
    else:
        template_ends = False
    if len(template_motif_1st_part)<ligation_window_length-(ligation_spot+1):
        template_begins = True
    else:
        template_begins = False

    if product_begins or not product_ends:
        product_category = motif_categories[-2-product_begins-product_ends]
    else:
        product_category = motif_categories[-1]

    if template_begins or not template_ends:
        template_category = motif_categories[-2-template_begins-template_ends]
    else:
        template_category = motif_categories[-1]
    return _production_channel_id(product_category, template_category,
            ligation_window_length, ligation_spot)

def _motif_production_reactants(ending_sequence : str,
        leaving_sequence : str,
        leaving_template_sequence : str,
        ending_template_sequence : str,
        motiflength : int,
        maximum_ligation_window_length : int
        ) -> Union[str,str,str,str,str]:
    """
    Returns:
    --------
    product_motif : str,
    ending_motif : str,
    leaving_motif : str,
    template_motif : str,
    production_channel_id : str
    """
    ending_motif = ending_sequence[-(motiflength-1):]
    leaving_motif = leaving_sequence[:(motiflength-1)]

    (ligation_window_length,
            ligation_spot,
            product_motif,
            ending_motif,
            leaving_motif,
            template_motif_1st_part,
            template_motif_2nd_part
            ) = _calculate_ligation_window_length_and_spot_and_motifs(
                    ending_sequence,
                    leaving_sequence,
                    leaving_template_sequence,
                    ending_template_sequence,
                    motiflength,
                    maximum_ligation_window_length
                    )
    production_channel_id = _determine_production_channel_id(
            ending_motif,
            leaving_motif,
            template_motif_1st_part,
            template_motif_2nd_part,
            ligation_window_length,
            ligation_spot,
            motiflength
            )

    template_motif = template_motif_1st_part+template_motif_2nd_part
    return product_motif, ending_motif, leaving_motif, template_motif, production_channel_id



from .fourmer_breakage_rates import fourmer_breakage_rates
from .fourmer_mass_correction import mass_correction_rates
from .fourmer_production_rates import _compute_extended_end_motif_reaction_logc_rates, _compute_extending_beginning_motif_reaction_logc_rates
from .fourmer_production_rates import _compute_produced_motif_reaction_logc_rates, compute_total_extension_rates
from .fourmer_production_rates import _initialize_empty_fourmer_production_rates
from .fourmer_trajectory_from_rate_constants import fourmer_trajectory_from_rate_constants
from .fourmer_trajectory_from_rate_constants import _fourmer_breakage_equations
from .fourmer_trajectory_from_rate_constants import _fourmer_production_equations
from .fourmer_trajectory_from_rate_constants import _fourmer_rate_equations
from .fourmer_trajectory_from_rate_constants import _integrate_motif_rate_equations

from .collisions import collisions_from_motif_concentration_trajectory_array_and_collision_rate_constants_array
from .collisions import motifs_collisions_array_from_motifs_array
from .collisions import ligation_spot_formations_from_motifs_array

from .dissociation_constant import (dissociation_constant_from_strand_reactor_parameters,
                                    ligation_rate_constant_from_strand_reactor_parameters,
                                    template_averaged_dissociation_constant_from_strand_reactor_parameters,
                                    energy_continuous_block,
                                    _energy_continuous_blocks,)


from .effective_ligation_rates_from_parameters import (effective_ligation_rates_from_parameters,
                                                       separate_hybridization_and_ligation_parameters
                                                       )

from .mean_length import mean_length

from .motif_concentration_vector import motif_concentration_vector_from_motif_number_vector
from .motif_concentration_trajectory import motif_concentration_trajectory_from_motif_number_trajectory
from .motif_production_rates import motif_production_rates_array_from_motif_production_rate_constants_array_and_motif_concentrations_array
from .motif_production_rates import motif_production_rates_array_from_motif_production_counts
from .motif_production_rate_constants_extension_matrix import motif_production_rate_constants_extension_matrix
from .motif_production_rate_constants_from_strand_reactor_parameters import motif_production_rate_constants_from_strand_reactor_parameters

from .motif_production_transition_kernel import _hybridization_site_categories
from .motif_production_transition_kernel import _extend_motif_production_rate_constants_array_to_collisions_format
from .motif_production_transition_kernel import motif_production_transition_kernel_from_motif_production_rate_constants_array
from .motif_production_transition_kernel import motif_production_transition_kernel_matrix

from .motif_breakage_rate_constants_from_strand_reactor_parameters import motif_breakage_rate_constants_from_strand_reactor_parameters

from .onset import onset_of_growth, discretized_onset_of_growth, onset_of_growth_from_strand_trajectory_parameters

from ._rates_utils import _set_invalid_logc_diff_to_zero, _set_invalid_logc_to_log0

from ._smooth_rate_clipping import _clip_smoothly, _clip_concentration_vector_smoothly

from .total_mass import total_mass, total_mass_of_motif_concentration_trajectory_array, total_mass_trajectory



from collections import namedtuple
import nifty8 as ift
import numpy as np

from ..utils.transform_vectors_to_fields import transform_dicts_to_field

def MotifBreakageTrajectory(motif_breakage_vectors : list,
                            times : ift.Field):
    motif_breakage_vector = motif_breakage_vectors[0]
    motif_breakages = [motif_breakage_vectors[ii].breakages for ii in range(len(motif_breakage_vectors))]
    motif_breakage_trajectory_properties = {'motiflength' : motif_breakage_vector.motiflength,
                                              'alphabet' : motif_breakage_vector.alphabet,
                                              'number_of_letters' : motif_breakage_vector.number_of_letters,
                                              'unit' : motif_breakage_vector.unit,
                                              'times': times,
                                              'breakages' : transform_dicts_to_field(motif_breakages, times.domain)
                                              }
    for motif_breakage_vector in motif_breakage_vectors:
        assert(motif_breakage_vector.motiflength==motif_breakage_trajectory_properties['motiflength'])
        assert(motif_breakage_vector.alphabet==motif_breakage_trajectory_properties['alphabet'])
        assert(motif_breakage_vector.number_of_letters==motif_breakage_trajectory_properties['number_of_letters'])
        assert motif_breakage_vector.unit==motif_breakage_trajectory_properties['unit'], "Units of motif breakage vectors need to be the same."
    mt = namedtuple('MotifBreakageTrajectory',
                tuple(motif_breakage_trajectory_properties.keys()))
    return mt(**motif_breakage_trajectory_properties)

def isinstance_motifbreakagetrajectory(obj) -> bool:
    is_motif_breakage_trajectory = True
    keys = ['motiflength', 'alphabet', 'number_of_letters', 'unit', 'times', 'breakages']
    for key in obj._asdict().keys():
        if key not in keys:
            print('Not a MotifBreakageTrajectory, missing key: {}.'.format(key))
            return False
    is_motif_breakage_trajectory *= isinstance(obj, tuple)
    is_motif_breakage_trajectory *= hasattr(obj, '_asdict')
    is_motif_breakage_trajectory *= hasattr(obj, '_fields')
    return bool(is_motif_breakage_trajectory)

def are_compatible_motif_breakage_trajectories(
        trj1 : MotifBreakageTrajectory,
        trj2 : MotifBreakageTrajectory
    ) -> bool:
    if not isinstance_motifbreakagetrajectory(trj1):
        print('Object is not a MotifBreakageTrajectory')
        return False
    if not isinstance_motifbreakagetrajectory(trj2):
        print('Object is not a MotifBreakageTrajectory')
        return False
    keys = ['motiflength', 'alphabet', 'unit', 'times']
    for key in keys:
        tt1 = trj1._asdict()[key]
        tt2 = trj2._asdict()[key]
        if isinstance(tt1, ift.Field):
            tt1 = tt1.val
        if isinstance(tt2, ift.Field):
            tt2 = tt2.val
        if not np.prod(tt1==tt2):
            print('MotifBreakageTrajectories not compatible: {} mismatch.'.format(key))
            return False
    return True



import morsaik as kdi
import numpy as np
from numpy.testing import assert_equal

def test__transform_sequence_array_to_motif_array():
    motiflengths = [4,5,6,7]
    number_of_letters = 4
    actual = np.ones((number_of_letters,)*4)
    for motiflength in motiflengths:
        desired = np.zeros((number_of_letters+1,)+(number_of_letters,)+(number_of_letters+1,)*(motiflength-2))
        if motiflength > 5:
            indices_actual = (0,slice(None),)+(slice(1,None),)*3+(0,)*(motiflength-5)
            desired[indices_actual] += actual
        elif motiflength == 5:
            desired[0,:,1:,1:,1:] += actual
        else:
            desired[1:,:,1:,1:] = actual
        assert_equal(kdi._transform_sequence_array_to_motif_array(actual,motiflength),
                desired)

def test__motif_vector_as_array():
    alphabet = ['a','b','c']
    motiflength = 5
    unit = kdi.make_unit('mol')/kdi.make_unit('L')

    number_of_letters = len(alphabet)
    motif_categories = kdi.domains._return_motif_categories(motiflength=motiflength)

    motif_number_dct = kdi._create_empty_motif_vector_dct(motiflength,alphabet=alphabet)
    desired = np.zeros(
            (number_of_letters+1,)
            +(number_of_letters,)
            +(number_of_letters+1,)*(motiflength-2)
            )
    for key in motif_number_dct.keys():
        current_array = np.random.randint(1e3)*np.ones(motif_number_dct[key].shape)
        current_indices = kdi._motif_indices_in_motifs_array(
                current_array,
                motiflength,
                key not in kdi.domains._motif_categories()[-2:]
                )
        motif_number_dct[key] = current_array
        desired[current_indices] = current_array

    motif_number_vector = kdi.MotifVector(motiflength,alphabet,unit)
    motif_number_vector = motif_number_vector(motif_number_dct)

    actual = kdi._motif_vector_as_array(motif_number_vector)
    assert_equal(actual,desired)


def test_save_and_load_motif_vector():
    archive_path = "./archive/test_motif_number_vector/"
    alphabet = ['a','b','c']
    motiflength = 5
    unit = kdi.make_unit('mol')/kdi.make_unit('L')

    motif_number_dct = kdi._create_empty_motif_vector_dct(motiflength,alphabet=alphabet)

    motif_number_vector = kdi.MotifVector(motiflength,alphabet,unit)
    motif_number_vector = motif_number_vector(motif_number_dct)

    kdi.save_motif_vector(archive_path, motif_number_vector)

    actual = kdi.load_motif_vector(archive_path)
    assert_equal(actual, motif_number_vector)
    assert(kdi.isinstance_motifvector(motif_number_vector))



from .strand_reactor_parameters import strand_reactor_parameters as get_strand_reactor_parameters
from .alphabet import alphabet as get_alphabet
from ..infer.motif_breakage_rate_constants_from_strand_reactor_parameters import motif_breakage_rate_constants_from_strand_reactor_parameters as infer_motif_breakage_rate_constants_from_strand_reactor_parameters 
from ..obj.motif_breakage_vector import MotifBreakageVector

def motif_breakage_rate_constants_from_strand_reactor_parameters(
        strand_trajectory_id : str,
        motiflength : int,
        complements : list,
        param_file_no : int = 0,
        standard_concentration : float = 1.
        ) -> MotifBreakageVector:
    return infer_motif_breakage_rate_constants_from_strand_reactor_parameters(
            get_strand_reactor_parameters(strand_trajectory_id, param_file_no),
            motiflength,
            get_alphabet(strand_trajectory_id),
            complements,
            standard_concentration
            )



from collections import namedtuple
import numpy as  np
import yaml
from os.path import exists
from scipy.sparse import coo_matrix, save_npz, load_npz

from .motif_production_trajectory import save_motif_production_trajectory
from .motif_production_trajectory import are_compatible_motif_production_trajectories
from .motif_production_trajectory import load_motif_production_trajectory

from ..utils.save import create_directory_path_if_not_already_existing

def MotifProductionTrajectoryEnsemble(motif_production_trajectories : list):
    """
    Ensemble of motif production trajectories
    """
    motif_production_trajectory = motif_production_trajectories[0]
    motif_production_trajectory_ensemble_properties = {
            'motiflength' : motif_production_trajectory.motiflength,
            'alphabet' : motif_production_trajectory.alphabet,
            'maximum_ligation_window_length' : motif_production_trajectory.maximum_ligation_window_length,
            'number_of_letters' : motif_production_trajectory.number_of_letters,
            'unit' : motif_production_trajectory.unit,
            'trajectories' : motif_production_trajectories
            }
    for motif_production_trajectory in motif_production_trajectories[1:]:
        assert(motif_production_trajectory.motiflength==motif_production_trajectory_ensemble_properties['motiflength'])
        assert(motif_production_trajectory.alphabet==motif_production_trajectory_ensemble_properties['alphabet'])
        assert(motif_production_trajectory.number_of_letters==motif_production_trajectory_ensemble_properties['number_of_letters'])
        assert motif_production_trajectory.unit==motif_production_trajectory_ensemble_properties['unit'], "Units of motif vectors need to be the same."
        assert(are_compatible_motif_production_trajectories(motif_production_trajectory,motif_production_trajectories[0]),
                "Motif trajectories not compatible.")
    mt = namedtuple('MotifProductionTrajectoryEnsemble',
                tuple(motif_production_trajectory_ensemble_properties.keys()))
    return mt(**motif_production_trajectory_ensemble_properties)

def isinstance_motifproductiontrajectoryensemble(obj) -> bool:
    is_motif_production_trajectory_ensemble = True
    keys = ['motiflength', 'alphabet', 'maximum_ligation_window_length', 'number_of_letters', 'unit', 'trajectories']
    for key in obj._asdict().keys():
        if key not in keys:
            print('Not a MotifProductionTrajectoryEnsemble, missing key: {}.'.format(key))
            return False
    is_motif_production_trajectory_ensemble *= isinstance(obj, tuple)
    is_motif_production_trajectory_ensemble *= hasattr(obj, '_asdict')
    is_motif_production_trajectory_ensemble *= hasattr(obj, '_fields')
    return bool(is_motif_production_trajectory_ensemble)

def are_compatible_motif_production_trajectory_ensembles(
        trjs1 : MotifProductionTrajectoryEnsemble,
        trjs2 : MotifProductionTrajectoryEnsemble
) -> bool:
    if not isinstance_motifproductiontrajectoryensemble(trjs1):
        print('Object is not a MotifProductionTrajectoryEnsemble')
        return False
    if not isinstance_motifproductiontrajectoryensemble(trjs2):
        print('Object is not a MotifProductionTrajectoryEnsemble')
        return False
    return np.prod([are_compatible_motif_production_trajectories(
        trj1,
        trj2
    ) for trj1 in trjs1.trajectories for trj2 in trjs2.trajectories])

def save_motif_production_trajectory_ensemble(
        archive_path : str,
        motif_production_trajectory_ensemble : MotifProductionTrajectoryEnsemble,
        file_space : bool = True
    ) -> None:
    create_directory_path_if_not_already_existing(archive_path)
    for trajectory_index in range(len(motif_production_trajectory_ensemble.trajectories)):
        current_path = archive_path+'motif_production_trajectory_{}/'.format(trajectory_index)
        create_directory_path_if_not_already_existing(current_path)
        save_motif_production_trajectory(current_path,
                motif_production_trajectory_ensemble.trajectories[trajectory_index]
                )
    with open(archive_path+'motif_production_trajectory_ensemble_properties.yaml','w') as yaml_file:
        yaml.dump({'motiflength':motif_production_trajectory_ensemble.motiflength,
            'alphabet':motif_production_trajectory_ensemble.alphabet,
            'maximum_ligation_window_length' : motif_production_trajectory_ensemble.maximum_ligation_window_length,
            'unit':motif_production_trajectory_ensemble.unit
                   },
            yaml_file,
            indent=4)

def load_motif_production_trajectory_ensemble(
        archive_path : str,
        file_sparse : bool = True
    ) -> MotifProductionTrajectoryEnsemble:
    dct_filename = archive_path+'motif_production_trajectory_ensemble_properties.yaml'
    motif_production_trajectories = []
    trajectory_index = 0
    while exists(archive_path+'motif_production_trajectory_{}/'.format(trajectory_index)):
        motif_production_trajectories = motif_production_trajectories + [load_motif_production_trajectory(
            archive_path+'motif_production_trajectory_{}/'.format(trajectory_index),
            file_sparse=file_sparse
        ),]
        trajectory_index += 1
    return MotifProductionTrajectoryEnsemble(motif_production_trajectories)



import morsaik as kdi
from numpy.testing import assert_equal

def test_transform_unit_to_dict_to_unit():
    q = kdi.make_unit('m')*kdi.make_unit('s')/kdi.make_unit('kg')
    qd = kdi.transform_unit_to_dict(q)
    actual = kdi.transform_dict_to_unit(qd)
    assert_equal(q,actual)



from ..read.strand_reactor_parameters import strand_reactor_parameters as read_strand_reactor_parameters
from ..utils.manage_strand_reactor_files import _create_typical_strand_parameters_filepath

def strand_reactor_parameters(
        strand_trajectory_id : str,
        param_file_no : int = 0
    ) -> dict:
    """
    gets strand reactor parameters for certain strand trajectory id

    Parameters:
    -----------
    strand_trajectory_id : str
        file path
    param_file_no : int

    Returns:
    --------
    strand_reactor_parameters : dict
        Dictionary with strand reactor parameters
    """
    filepath = _create_typical_strand_parameters_filepath(strand_trajectory_id,
            param_file_no=param_file_no)
    return read_strand_reactor_parameters(filepath)



import numpy as np
import jax.numpy as jnp
import yaml
from ..domains.motif_space import _return_motif_categories
from collections import namedtuple
import nifty8 as ift

from .units import make_unit
from .units import transform_unit_to_dict, transform_dict_to_unit

from ..domains.motif_space import MotifSpace
from ..domains.hamming_space import HammingSpace

from ..utils.save import create_directory_path_if_not_already_existing

def MotifVector(motiflength : int,
        alphabet : list,
        unit : str):
    motif_categories = _return_motif_categories(motiflength=motiflength)
    motif_vector_properties = {'motiflength' : motiflength,
        'alphabet' : alphabet,
        'number_of_letters' : len(alphabet),
        'unit' : make_unit(unit)}
    def makeMotifVector(motif_vector_dct : dict):
        motif_vector = namedtuple('MotifVector',
                ('motifs',) + tuple(motif_vector_properties.keys()))
        motifs = ift.MultiField.from_dict(
            _motif_vector_dct_with_fields(motif_vector_dct, alphabet),
            MotifSpace.make(alphabet, motiflength, units=make_unit('bits')),
        )
        return motif_vector(**{**{'motifs': motifs}, **motif_vector_properties})
    return makeMotifVector

def add_zebra_fluctuation_to_motif_vector(motif_vector : MotifVector,
                                          zebra_fluctuation : float = 0.,
                                          braze_fluctuation : float = 0.,
                                          aa_fluctuation : float = 0.,
                                          bb_fluctuation : float = 0.,
                                          fourmer_fluctuation : float = 0.
    ) -> MotifVector:
    motif_array = _motif_vector_as_array(motif_vector)
    ic = motif_array[0,0,2,0]
    motif_array[0,0,2,0] += zebra_fluctuation*motif_array[0,0,2,0]
    motif_array[0,1,1,0] += braze_fluctuation*motif_array[0,1,1,0]

    motif_array[0,0,1,0] += aa_fluctuation*motif_array[0,0,1,0]
    motif_array[0,1,2,0] += bb_fluctuation*motif_array[0,1,2,0]

    motif_array[0,0,2,1] += fourmer_fluctuation*ic
    motif_array[1,1,1,2] += fourmer_fluctuation*ic
    motif_array[2,0,2,0] += fourmer_fluctuation*ic
    return MotifVector(motif_vector.motiflength, motif_vector.alphabet, motif_vector.unit)(_array_to_motif_vector_dct(motif_array, motif_vector.motiflength, motif_vector.alphabet))

def convert_homogeneous_dimers_to_zebra_dimers(
        motif_vector : MotifVector,
        zebra_dimer_concentration : float
    ) -> MotifVector:
    """
    Note: if zebra_dimer_concentration is negative, homogeneous dimers are added and zebra_dimers are reduced.
    """
    motif_array = _motif_vector_as_array(motif_vector)
    motif_array[0,0,2,0] += zebra_dimer_concentration
    motif_array[0,1,1,0] += zebra_dimer_concentration
    motif_array[0,1,2,0] -= zebra_dimer_concentration
    motif_array[0,0,1,0] -= zebra_dimer_concentration
    return MotifVector(motif_vector.motiflength, motif_vector.alphabet, motif_vector.unit)(_array_to_motif_vector_dct(motif_array, motif_vector.motiflength, motif_vector.alphabet))

def convert_dimers_to_tetramers(
        motif_vector : MotifVector,
        zebra_tetramer_concentration : float
    ) -> MotifVector:
    """
    Note: if zebra_tetramer_concentration is negative, homogeneous dimers are converted to tetramers instead.
    """
    motif_array = _motif_vector_as_array(motif_vector)

    if zebra_tetramer_concentration >= 0.:
        motif_array[0,0,2,1] += .5*zebra_tetramer_concentration
        motif_array[1,1,1,2] += .5*zebra_tetramer_concentration
        motif_array[2,0,2,0] += .5*zebra_tetramer_concentration

        motif_array[0,1,1,2] += .5*zebra_tetramer_concentration
        motif_array[2,0,2,1] += .5*zebra_tetramer_concentration
        motif_array[1,1,1,0] += .5*zebra_tetramer_concentration

        motif_array[0,0,2,0] -= zebra_tetramer_concentration
        motif_array[0,1,1,0] -= zebra_tetramer_concentration

    else:
        motif_array[0,0,1,1] += .5*zebra_tetramer_concentration
        motif_array[1,0,1,1] += .5*zebra_tetramer_concentration
        motif_array[1,0,1,0] += .5*zebra_tetramer_concentration

        motif_array[0,1,2,2] += .5*zebra_tetramer_concentration
        motif_array[2,1,2,2] += .5*zebra_tetramer_concentration
        motif_array[2,1,2,0] += .5*zebra_tetramer_concentration

        motif_array[0,1,2,0] -= zebra_tetramer_concentration
        motif_array[0,0,1,0] -= zebra_tetramer_concentration

    return MotifVector(motif_vector.motiflength, motif_vector.alphabet, motif_vector.unit)(_array_to_motif_vector_dct(motif_array, motif_vector.motiflength, motif_vector.alphabet))

def _motif_vector_dct_with_fields(motif_vector_dct,
                                  alphabet : list) -> dict:
    for key in motif_vector_dct.keys():
        if not isinstance(motif_vector_dct[key],ift.Field):
            motif_vector_dct[key] = ift.Field(
                ift.DomainTuple.make(HammingSpace(alphabet,len(motif_vector_dct[key].shape))),
                np.asarray(motif_vector_dct[key]))
    return motif_vector_dct

def _create_empty_motif_vector_dct(motiflength : int,
        alphabet : list = ['a','b']) -> dict:
    number_of_letters = len(alphabet)

    empty_motif_vector = {}
    for strandlength in range(1,motiflength-1):
        category = _return_motif_categories()[0].format(strandlength)
        shape = (number_of_letters,)*strandlength
        empty_motif_vector[category] = np.zeros(shape)
    motif_categories = _return_motif_categories(motiflength=motiflength)
    for category in motif_categories[-3:]:
        category_length = motiflength-1 + (category==motif_categories[-2])
        shape = (number_of_letters,)*category_length
        empty_motif_vector[category] = np.zeros(shape)
    return empty_motif_vector

def isinstance_motifvector(obj : object,
                           print_statements : bool = True
                           ) -> bool:
    is_motif_vector = True
    keys = ['motifs','motiflength', 'alphabet', 'number_of_letters', 'unit']
    for key in obj._asdict().keys():
        if key not in keys:
            if print_statements:
                print('Not a MotifVector, missing key: {}.'.format(key))
            return False
    keys = list(_return_motif_categories(motiflength=obj.motiflength))
    for key in obj.motifs.keys():
        if key not in keys:
            if print_statements:
                print('Not a MotifVector, missing key in motifs field: {}.'.format(key))
            return False
    is_motif_vector *= isinstance(obj, tuple)
    is_motif_vector *= hasattr(obj, '_asdict')
    is_motif_vector *= hasattr(obj, '_fields')
    return bool(is_motif_vector)

def are_compatible_motif_vectors(mv1 : MotifVector, mv2 : MotifVector) -> bool:
    if not isinstance_motifvector(mv1):
        print('Object is not a MotifVector')
        return False
    if not isinstance_motifvector(mv2):
        print('Object is not a MotifVector')
        return False
    keys = ['motiflength', 'alphabet', 'unit']
    for key in keys:
        if not np.prod(mv1._asdict()[key]==mv2._asdict()[key]):
            print('MotifVectors not compatible: {} mismatch.'.format(key))
            return False
    return True

def _motif_indices_in_motifs_array(motif_vector_array : np.ndarray,
        motiflength : int,
        is_beginning : bool
        ) -> tuple:
    """
    returns an index-tuple for the array in which the motif_sequence_array fits.

    Parameters:
    -----------
    motif_vector_array : np.ndarray,
    motiflength : int,
    is_beginning : bool

    Returns:
    --------
    motif_indices : tuple
    """
    if len(motif_vector_array.shape)==motiflength:
        is_beginning = False
    strandlength = len(motif_vector_array.shape)
    number_of_letters = motif_vector_array.shape[0]
    current_indices = (0,)*int(is_beginning)
    current_indices += (slice(1,number_of_letters+1),)*int(not is_beginning)
    current_indices += (slice(0,number_of_letters),)
    current_indices += (slice(1,number_of_letters+1),)*int(
            strandlength
            -1
            -int(not is_beginning)
            )
    current_indices += (0,)*(motiflength-len(current_indices))
    return current_indices

def categories_indices(motiflength : int,
                       alphabet : list
                       ) -> dict:
    """
    returns a dictionary with motif categories as keys and index-tuples as their values.

    Parameters:
    -----------
    motiflenth : int
    alphabet : list

    Returns:
    --------
    categories_indices : dict
    """
    keys = _return_motif_categories(motiflength)

    number_of_letters = len(alphabet)
    categories_indices = {}
    for strandlength in range(1,motiflength-1):
        is_beginning = True
        key = keys[strandlength-1]

        current_indices = (0,)*int(is_beginning)
        current_indices += (slice(1,number_of_letters+1),)*int(not is_beginning)
        current_indices += (slice(0,number_of_letters),)
        current_indices += (slice(1,number_of_letters+1),)*int(
                strandlength
                -1
                -int(not is_beginning)
                )
        current_indices += (0,)*(motiflength-len(current_indices))

        categories_indices[key] = current_indices

    for key in keys[-3:]:
        is_beginning = (key==keys[-3])
        strandlength = motiflength-int(key!=keys[-2])

        current_indices = (0,)*int(is_beginning)
        current_indices += (slice(1,number_of_letters+1),)*int(not is_beginning)
        current_indices += (slice(0,number_of_letters),)*(motiflength>2)
        current_indices += (slice(1,number_of_letters+1),)*int(
                strandlength
                -int(motiflength>2)
                -int(not is_beginning)
                )
        current_indices += (0,)*(motiflength-len(current_indices))

        categories_indices[key] = current_indices
    return categories_indices

def category_indices(motif_category : str,
                     motiflength : int,
                     alphabet : list
                     ) -> tuple:
    """
    returns the array-indix-tuple for a certain motif_category.
    """
    return categories_indices(motiflength, alphabet)[motif_category]

def _transform_sequence_array_to_motif_array(
        sequence_array,
        motiflength : int,
        sequence_is_beginning : bool = True
        ) -> np.ndarray:
    """
    transforms an array that only tracks letters (occupied nucleotides)
    into an array that explicitely tracks empty spots with zeros.
    Note that the second spot is always a letter, though.

    Parameters:
    -----------
    sequence_array : nd-array
    motiflength : int
    sequence_is_beginning : boolean (optional)
        whether the sequence is beginning, i.e. the first spot is zero,
        else it will be considered as end or continuation if all spots are
        letters
        default : True

    Returns:
    --------
    motif_array : nd-array
    """
    if len(sequence_array.shape)==motiflength:
        sequence_is_beginning = False
    number_of_letters = sequence_array.shape[0]
    strandlength = len(sequence_array.shape)

    motif_array_shape = (number_of_letters+1,)
    motif_array_shape += (number_of_letters,)*int(motiflength>2)
    motif_array_shape += (number_of_letters+1,)*(motiflength-1-int(motiflength>2))
    motif_array = np.zeros(motif_array_shape)

    indices_sequence_arrays = (0,)*int(sequence_is_beginning)
    indices_sequence_arrays += (slice(1,None),)*(1-int(sequence_is_beginning))
    indices_sequence_arrays += (slice(None),)*int(motiflength>2)
    indices_sequence_arrays += (slice(1,None),)*(strandlength-1-int(motiflength>2)+int(sequence_is_beginning))
    indices_sequence_arrays += (0,)*(motiflength-len(indices_sequence_arrays))

    motif_array[indices_sequence_arrays] = sequence_array
    return motif_array

def _motif_vector_as_array(motif_vector : MotifVector
        ) -> np.ndarray:
    """
    transforms a motif vector into an numpy-array

    Parameters:
    -----------
    motif_vector : MotifVector

    Returns:
    --------
    motif_array : np.ndarray
    """
    motiflength = motif_vector.motiflength
    number_of_letters = motif_vector.number_of_letters

    motif_categories = _return_motif_categories(motiflength)

    motif_array_shape = (number_of_letters+1,)
    motif_array_shape += (number_of_letters,)*int(motiflength>2)
    motif_array_shape += (number_of_letters+1,)*(motiflength-1-int(motiflength>2))
    motif_array = np.zeros(motif_array_shape)
    for motif_category in motif_categories[:-2]:
        motif_array += _transform_sequence_array_to_motif_array(
                motif_vector.motifs[motif_category].val,
                motiflength,
                sequence_is_beginning = True
                )
    for motif_category in motif_categories[-2:]:
        motif_array += _transform_sequence_array_to_motif_array(
                motif_vector.motifs[motif_category].val,
                motiflength,
                sequence_is_beginning = False
                )
    return motif_array

def _array_to_motif_vector_dct(motif_vector_array : np.ndarray,
        motiflength : int,
        alphabet : list,
        ) -> dict:
    if motiflength != len(motif_vector_array.shape):
        raise ValueError("motiflength inconsistent with array shape")
    number_of_letters = len(alphabet)
    if number_of_letters != motif_vector_array.shape[0]-1:
        raise ValueError("alphabet shape ({alphabet_shape}) inconsistent with motif_vector_array shape ({motif_vector_array_shape})".format(
            alphabet_shape = len(alphabet),
            motif_vector_array_shape = motif_vector_array.shape
                    ))

    motif_vector_dct = {}
    for strandlength in range(1,motiflength):
        if strandlength == (motiflength-1):
            category = _return_motif_categories()[-3]
        else:
            category = _return_motif_categories()[0].format(strandlength)
        current_indices = (0,)
        current_indices += (slice(None),)*int(motiflength>2)
        current_indices += (slice(1,None),)*(strandlength-int(motiflength>2))
        current_indices += (0,)*(motiflength-len(current_indices))
        motif_vector_dct[category] = motif_vector_array[current_indices]
    motif_categories = _return_motif_categories(motiflength=motiflength)
    for category in motif_categories[-2:]:
        strandlength = motiflength-1 + (category==motif_categories[-2])
        current_indices = (slice(1,None),)
        current_indices += (slice(None),)*int(motiflength>2)
        current_indices += (slice(1,None),)*(motiflength-2-int(motiflength>2)+int(category==motif_categories[-2]))
        current_indices += (0,)*(motiflength-len(current_indices))
        motif_vector_dct[category] = motif_vector_array[current_indices]
    return motif_vector_dct

def save_motif_vector(archive_path : str,
        motif_vector : MotifVector
        ) -> None:
    create_directory_path_if_not_already_existing(archive_path)
    jnp.save(archive_path+'motifs',
            _motif_vector_as_array(motif_vector)
            )

    with open(archive_path+'properties.yaml','w') as yaml_file:
        yaml.dump({'motiflength':motif_vector.motiflength,
            'alphabet':motif_vector.alphabet,
            'unit':transform_unit_to_dict(motif_vector.unit)},
            yaml_file,
            indent=4)

def load_motif_vector(archive_path : str
        ) -> MotifVector:
    dct_filename =  archive_path+'properties'+'.yaml'
    array_filename = archive_path+'motifs'+'.npy'
    with open(dct_filename, 'r') as yaml_file:
        motif_vector_properties = yaml.safe_load(yaml_file)
    motif_vector_properties['unit'] = transform_dict_to_unit(motif_vector_properties['unit'])
    makeMotifVector = MotifVector(**motif_vector_properties)
    return makeMotifVector(
            _array_to_motif_vector_dct(jnp.load(array_filename),motif_vector_properties['motiflength'], motif_vector_properties['alphabet'])
            )



import jax.numpy as jnp
from warnings import warn

from jax import config
config.update("jax_enable_x64", True)

from ._rates_utils import _set_invalid_logc_to_log0
from ._rates_utils import _set_invalid_logc_diff_to_zero
from ._rates_utils import _cut_low_concentrations
from ._smooth_rate_clipping import _clip_concentration_vector_smoothly

def _cut_small_rates(rates, rates_cut):
    # logc_diff = jnp.where((fourmer_logc<=jnp.log(pseudo_count_concentration))*(logc_diff<0.),0.,logc_diff)
    return jnp.where(
            rates<rates_cut,
            0.,
            jnp.sqrt(rates**2-rates_cut**2)
            )

def _left_continuation_frequency(
        fourmer_logc : jnp.ndarray,
        ) -> jnp.ndarray:
    return jnp.exp(fourmer_logc)/(jnp.sum(jnp.exp(fourmer_logc),axis=0)[None])

def _right_continuation_frequency(
        fourmer_logc : jnp.ndarray,
        ) -> jnp.ndarray:
    return jnp.exp(fourmer_logc)/(jnp.sum(jnp.exp(fourmer_logc),axis=-1)[:,:,:,None])

def _effective_left_breakage_frequency(
        fourmer_logc : jnp.ndarray,
        effective_breakage_rate_constants : jnp.ndarray,
        ) -> jnp.ndarray:
    left_continuation_frequency = _left_continuation_frequency(fourmer_logc)
    return jnp.sum(left_continuation_frequency*effective_breakage_rate_constants,axis=0)[:,:,:,None]

def _effective_central_breakage_frequency(
        effective_breakage_rate_constants : jnp.ndarray,
        ) -> jnp.ndarray:
    return effective_breakage_rate_constants

def _effective_right_breakage_frequency(
        fourmer_logc : jnp.ndarray,
        effective_breakage_rate_constants : jnp.ndarray
        ) -> jnp.ndarray:
    right_continuation_frequency = _right_continuation_frequency(fourmer_logc)
    return jnp.sum(right_continuation_frequency*effective_breakage_rate_constants,axis=-1)[None]

def _break_central_bonds(
        logc_diff : jnp.ndarray,
        effective_breakage_rate_constants : jnp.ndarray,
        pseudo_count_concentration : float = 1.e-12
        ) -> jnp.ndarray:
    effective_central_breakage_frequency = _effective_central_breakage_frequency(effective_breakage_rate_constants)
    logc_diff = logc_diff.at[:,1:,1:,:].subtract(effective_central_breakage_frequency[:,1:,1:,:])
    return _set_invalid_logc_to_log0(logc_diff, pseudo_count_concentration=pseudo_count_concentration)

def _break_left_bonds(
        logc_diff : jnp.ndarray,
        fourmer_logc : jnp.ndarray,
        effective_breakage_rate_constants : jnp.ndarray,
        pseudo_count_concentration : float = 1.e-12
        ) -> jnp.ndarray:
    effective_left_breakage_frequency = _effective_left_breakage_frequency(
            fourmer_logc,
            effective_breakage_rate_constants
            )
    logc_diff = logc_diff.at[1:,1:,1:,:].subtract(effective_left_breakage_frequency[1:,1:,1:,:])
    return _set_invalid_logc_to_log0(logc_diff, pseudo_count_concentration=pseudo_count_concentration)

def _break_right_bonds(
        logc_diff : jnp.ndarray,
        fourmer_logc : jnp.ndarray,
        effective_breakage_rate_constants : jnp.ndarray,
        pseudo_count_concentration : float = 1.e-12
        ) -> jnp.ndarray:
    effective_right_breakage_frequency = _effective_right_breakage_frequency(
            fourmer_logc,
            effective_breakage_rate_constants)
    logc_diff = logc_diff.at[:,1:,1:,1:].subtract(effective_right_breakage_frequency[:,1:,1:,1:])
    return _set_invalid_logc_to_log0(logc_diff, pseudo_count_concentration=pseudo_count_concentration)

def _produce_monomers(
        logc_diff : jnp.ndarray,
        fourmer_logc : jnp.ndarray,
        effective_breakage_rate_constants : jnp.ndarray,
        pseudo_count_concentration : float = 1.e-12
        ) -> jnp.ndarray:
    effective_central_breakage_frequency = _effective_central_breakage_frequency(effective_breakage_rate_constants)
    logc_diff = logc_diff.at[0,1:,0,0].add(jnp.sum(
            effective_central_breakage_frequency[0,1:,1:,:]*jnp.exp(fourmer_logc[0,1:,1:,:]-fourmer_logc[0,1:,0,0][:,None,None]),
            axis=(-1,-2)
            ))
    logc_diff = _set_invalid_logc_to_log0(logc_diff)
    logc_diff = logc_diff.at[0,1:,0,0].add(jnp.sum(
            effective_central_breakage_frequency[:,1:,1:,0]*jnp.exp(fourmer_logc[:,1:,1:,0]-fourmer_logc[0,1:,0,0][None,None,:]),
            axis=(0,1)
            ))
    return _set_invalid_logc_to_log0(logc_diff)

def _produce_ends(
        logc_diff : jnp.ndarray,
        fourmer_logc : jnp.ndarray,
        effective_breakage_rate_constants : jnp.ndarray,
        pseudo_count_concentration : float = 1.e-12
        ) -> jnp.ndarray:
    effective_right_breakage_frequency = _effective_right_breakage_frequency(
            fourmer_logc,
            effective_breakage_rate_constants
            )
    logc_diff = logc_diff.at[:,1:,1:,0].add(jnp.sum(
            effective_right_breakage_frequency[:,1:,1:,1:]*jnp.exp(fourmer_logc[:,1:,1:,1:]-fourmer_logc[:,1:,1:,0][:,:,:,None]),
            axis=-1
            ))
    return _set_invalid_logc_to_log0(logc_diff)

def _produce_beginnings(
        logc_diff : jnp.ndarray,
        fourmer_logc : jnp.ndarray,
        effective_breakage_rate_constants : jnp.ndarray,
        pseudo_count_concentration : float = 1.e-12
        ) -> jnp.ndarray:
    effective_left_breakage_frequency = _effective_left_breakage_frequency(
            fourmer_logc,
            effective_breakage_rate_constants
            )
    logc_diff = logc_diff.at[0,1:,1:,:].add(
            jnp.sum(
                effective_left_breakage_frequency[1:,1:,1:,:]*jnp.exp(fourmer_logc[1:,1:,1:,:]-fourmer_logc[0,1:,1:,:][None]),
                axis=0
                )
            )
    return _set_invalid_logc_to_log0(logc_diff)

def fourmer_breakage_rates(
        fourmer_logc : jnp.ndarray,
        effective_breakage_rate_constants : jnp.ndarray = 0.,
        pseudo_count_concentration : float = 1.e-12,
        fourmer_logc_slope : float = 0.,
        logc_diff_slope : float = 0.,
        soft_reactant_threshold : float = None,
        hard_reactant_threshold : float = None,
        ) -> jnp.ndarray:
    """
    Performs the breakage of the left breaking motif at the breaking point
    (1,2,3,4) three breaking points
    (1,2,3,0) two breaking points
    (0,2,3,4) two breaking points
    (0,2,3,0) one breaking point
    (0,2,0,0) no breaking point
    """
    if soft_reactant_threshold is None:
        soft_reactant_threshold = pseudo_count_concentration
    if hard_reactant_threshold is None:
        hard_reactant_threshold = soft_reactant_threshold/2.
    if isinstance(effective_breakage_rate_constants, (float,int)):
        if effective_breakage_rate_constants == 0.:
            return 0.
        effective_breakage_rate_constants = effective_breakage_rate_constants*jnp.ones(fourmer_logc.shape)
        effective_breakage_rate_constants = _set_invalid_logc_diff_to_zero(
                effective_breakage_rate_constants,
                print_warning = False
                )
    cut_off_weight = _clip_concentration_vector_smoothly(jnp.exp(fourmer_logc), soft_reactant_threshold, hard_reactant_threshold)
    effective_breakage_rate_constants = cut_off_weight*effective_breakage_rate_constants
    logc_diff = jnp.zeros(fourmer_logc.shape)
    logc_diff = _break_central_bonds(
            logc_diff,
            effective_breakage_rate_constants,
            pseudo_count_concentration=pseudo_count_concentration
            )
    logc_diff = _break_left_bonds(
            logc_diff,
            fourmer_logc,
            effective_breakage_rate_constants,
            pseudo_count_concentration=pseudo_count_concentration
            )
    logc_diff = _break_right_bonds(
            logc_diff,
            fourmer_logc,
            effective_breakage_rate_constants,
            pseudo_count_concentration=pseudo_count_concentration
            )
    logc_diff = _produce_ends(
            logc_diff,
            fourmer_logc,
            effective_breakage_rate_constants,
            pseudo_count_concentration=pseudo_count_concentration
            )
    logc_diff = _produce_beginnings(
            logc_diff,
            fourmer_logc,
            effective_breakage_rate_constants,
            pseudo_count_concentration=pseudo_count_concentration
            )
    logc_diff = _produce_monomers(
            logc_diff,
            fourmer_logc,
            effective_breakage_rate_constants,
            pseudo_count_concentration=pseudo_count_concentration
            )
    return _set_invalid_logc_diff_to_zero(logc_diff, print_warning = False)

def _shape_brc(
        breakage_rate_constants : jnp.ndarray,
        number_of_letters : int = 4,
        motiflength : int = 4 
        ) -> jnp.ndarray:
    # breakage_rate_constants_shape = # # (number_of_letters+1,)*(motiflength//2-1)+(number_of_letters,number_of_letters,)+(number_of_letters+1,)*(motiflength//2-1)
    breakage_rate_constants_shape = (number_of_letters+1,)*motiflength
    if isinstance(breakage_rate_constants,float):
        brc = jnp.zeros(breakage_rate_constants_shape)
        breakage_rate_constants = brc.at[:,1:,1:,:].add(breakage_rate_constants)
    elif (jnp.prod(jnp.asarray(breakage_rate_constants.shape))==(number_of_letters+1)**2*number_of_letters**2):
        brc = jnp.zeros(breakage_rate_constants_shape)
        breakage_rate_constants = brc.at[:,1:,1:,:].add(breakage_rate_constants)
    elif breakage_rate_constants.shape != breakage_rate_constants_shape:
        raise ValueError(f"{breakage_rate_constants.shape=} != {breakage_rate_constants_shape=}")
    return breakage_rate_constants



import morsaik as kdi
import numpy as np
from numpy.testing import assert_equal
import pytest

def test_motif_production_trajectory():
    alphabet = ['a','b','c']
    motiflength = 5
    unit = kdi.make_unit('mol')/kdi.make_unit('L')
    times = kdi.TimesVector(np.arange(3),'ms')
    maximum_ligation_window_length = motiflength

    motif_production_vector_dct = kdi._create_empty_motif_production_dict(
            motiflength,
            alphabet,
            maximum_ligation_window_length
            )
    motif_production_vector = kdi.MotifProductionVector(motiflength,alphabet,unit,
            maximum_ligation_window_length)
    motif_production_vector = motif_production_vector(motif_production_vector_dct)
    motif_production_vectors = [motif_production_vector,]*times.shape[0]
    print('askdfjghasdf')
    print(motif_production_vectors)
    print(times)
    mpt = kdi.MotifProductionTrajectory(motif_production_vectors, times)

    assert_equal(mpt.times.val,times.val)
    assert_equal(mpt.times.domain[0].units,times.domain[0].units)
    assert(mpt.alphabet==alphabet)
    assert(mpt.number_of_letters==len(alphabet))
    assert(mpt.motiflength==motiflength)
    assert(mpt.unit==unit)
    for key in mpt.productions.val.keys():
        assert_equal(mpt.productions[key].val,
                np.asarray([motif_production_vectors[ii].productions[key].val for ii in range(times.shape[0])])
                )

def test_motif_producion_trajectory():
    alphabet = ['a','b','c']
    motiflength = 5
    unit = 'mol/L'
    times = kdi.TimesVector(np.arange(3),'ms')
    maximum_ligation_window_length = motiflength

    motif_production_vector_dct = kdi._create_empty_motif_production_dict(
            motiflength,
            alphabet,
            maximum_ligation_window_length)

    motif_production_vector = kdi.MotifProductionVector(motiflength,alphabet,unit,
            maximum_ligation_window_length)
    motif_production_vector2 = kdi.MotifProductionVector(motiflength,alphabet,'a',
            maximum_ligation_window_length)
    motif_production_vector = motif_production_vector(motif_production_vector_dct)
    motif_production_vector2 = motif_production_vector2(motif_production_vector_dct)
    motif_production_vectors = [motif_production_vector,]*(times.shape[0]-1)
    motif_production_vectors += [motif_production_vector2,]
    with pytest.raises(AssertionError):
        mpt = kdi.MotifProductionTrajectory(motif_production_vectors, times)

def test_save_and_load_motif_production_trajectory():
    motiflengths = [4,] # TODO: [2,3,4,5,6,10]
    for motiflength in motiflengths:
        strand_trajectory_id='9999_99_99__99_99_99'
        archive_path = './archive/{strand_trajectory_id}/'.format(strand_trajectory_id=strand_trajectory_id)
        motif_production_trajectory = kdi.get.strand_motifs_productions_trajectory(
            motiflength,
            strand_trajectory_id,
        )
        kdi.save_motif_production_trajectory(archive_path, motif_production_trajectory)
        actual = kdi._motif_production_trajectory_as_array(
            kdi.load_motif_production_trajectory(archive_path)
        )
        assert_equal(actual,
                     kdi._motif_production_trajectory_as_array(motif_production_trajectory)
                     )



from .config import config, symbol_config

from .strand_reactor_parameters import strand_reactor_parameters
from .strand_reactor_nucleotides import strand_reactor_nucleotides

from .strand_motifs_trajectory import strand_motifs_trajectory, steps_and_times_and_complexes_from_complexes_txt
from .strand_motifs_trajectory import _extract_upper_and_lower_strands_as_continuous_strings_from_structure_of_complex
from .strand_motifs_trajectory import _cut_strands_as_continuous_string_into_separate_strand_sequences
from .strand_motifs_trajectory import steps_and_times_and_sequence_trajectory_from_complexes_txt
from .strand_motifs_trajectory import _transform_motif_string_to_index_tuple, _transform_sequence_vector_into_motif_vector
from .strand_motifs_trajectory_ensemble import strand_motifs_trajectory_ensemble

from .strand_motifs_productions_trajectory import strand_motifs_productions_trajectory
from .strand_motifs_productions_trajectory import _ligation_statistics, _step_simulation_and_ligation_strands_array
from .strand_motifs_productions_trajectory import _motif_production_reactants
from .strand_motifs_productions_trajectory import _ligation_spot, _determine_production_channel_id

from .strand_motifs_productions_trajectory_ensemble import strand_motifs_productions_trajectory_ensemble

from .strand_length_distribution import strand_length_distribution



import jax.numpy as jnp
import jax
from itertools import product as iterprod
from jax.experimental import sparse

def _hybridization_site_categories(
        number_of_letters : int = 2,
        hybridization_length_max : int = 4
        ) -> list[tuple[int]]:
    """
    returns all possible hybridization site categories for a given
    number of letters and maximum hybridization length
    in terms of the lenght of the hybridization site, the length of the left, right and template reactant and the shift (overhang) of the left strand vs the template strand.

    Parameters:
    -----------
    hybridization_length_max : int

    Returns:
    --------
    hybridization_site_categories : list[tuple[int]]
        hybridization_length, left_ligant_length, right_ligant_length, template_length, shift_length
    """
    hybridization_site_categories = []
    for hybridization_length in range(2,hybridization_length_max+1):
        for left_ligant_length in range(1, hybridization_length):
            for right_ligant_length in range(max(1,hybridization_length-left_ligant_length-(hybridization_length>2)-(hybridization_length>3)),hybridization_length-left_ligant_length+1):
                for template_length in range(max(hybridization_length-(hybridization_length>2)-(hybridization_length>3),2*hybridization_length-(hybridization_length>2)-(hybridization_length>3)-right_ligant_length-left_ligant_length),hybridization_length+1):
                    for shift_length in range(-1,2):
                        if left_ligant_length==1 and shift_length <0:
                            continue
                        if right_ligant_length==1 and shift_length+right_ligant_length+left_ligant_length-template_length>0:
                            continue
                        if (hybridization_length==(left_ligant_length+right_ligant_length+(shift_length==1)+(shift_length==template_length-right_ligant_length-left_ligant_length-1))) and (hybridization_length==template_length+(shift_length==-1)+(shift_length+right_ligant_length+left_ligant_length-template_length==1)):
                            hybridization_site_categories += [(hybridization_length,left_ligant_length,right_ligant_length,template_length,shift_length)]
    number_hybridization_configurations = [number_of_letters**(hybridization_site_category[1]+hybridization_site_category[2]+hybridization_site_category[3]) for hybridization_site_category in hybridization_site_categories]
    hybridization_configuration_indices = jnp.cumsum(jnp.array([0]+number_hybridization_configurations))
    return hybridization_site_categories, hybridization_configuration_indices

def print_hybridization_site_categories(hybridization_length_max : int = 4):
    hybridization_site_categories = _hybridization_site_categories(hybridization_length_max)
    for hybridization_site_category in hybridization_site_categories:
        print('\n'+'0'*hybridization_site_category[0]+f"{hybridization_site_category[0]}")
        print('0'*(hybridization_site_category[-1]==1)+'L'*hybridization_site_category[1]+'|'+'R'*hybridization_site_category[2])
        print('0'*(hybridization_site_category[-1]==-1)+'T'*(hybridization_site_category[1]+a[-1])+'-'+'T'*(hybridization_site_category[3]-hybridization_site_category[1]-hybridization_site_category[-1]))
        print(hybridization_site_category)

def _extend_motif_production_rate_constants_array_to_collisions_format(
        motif_production_rate_constants : jax.Array,
        maximum_ligation_window_length : int = 4,
        motiflength : int = 4,
        number_of_letters : int = 4
        ) -> jax.Array:
    """
    The motif_production_rate_constants are categorized by the ligation windows.
    For the transition kernal they first need to be extended to the shape of the ligation_sites,
    which means a degeneracy, i.e. all ligation_sites with the same ligation window get the same contribution.

    In principle, there are the following different ligation windows:
    ((0,0),(0,0)); ((0,1:),(0,0)); ((1:,0),(0,0)); ((0,0),(0,1:));((0,0),(1:,0));((0,0),(1:,1:));((1:,0),(1:,0));((0,1:),(0,1:)),((1:,1:),(0,0))
    for ligation_window_length<=maximum_ligation_window_length
    Ligation windows of ligation_window_length=maximum_ligation_window_length, we additionally have
    ((1:,1:),(0,1:));((1:,1:),(1:,0));((1:,0),(1:,1:);((0,1:),(1:,1:));((1:,1:),(1:,1:))
    """
    if maximum_ligation_window_length <4:
        ligation_window_lengths = [1]
        raise NotImplementedError("Only implemented for ligation window length greater than 4 so far.")
    else:
        ligation_window_lengths = range(4,maximum_ligation_window_length+1)
    number_of_motifs = jnp.concatenate((number_of_letters**jnp.arange(1,motiflength+1),jnp.array([number_of_letters**(motiflength-1)])))
    number_of_strands = number_of_motifs[:-3]
    number_of_beginnings = number_of_motifs[-3]
    number_of_continuations = number_of_motifs[-2]
    number_of_ends = number_of_motifs[-1]

    reactant_indices = jnp.concatenate((jnp.zeros(1),jnp.cumsum(number_of_letters**jnp.arange(1,motiflength+1))), dtype=int)
    template_indices = jnp.concatenate((jnp.zeros(2),jnp.cumsum(jnp.concatenate((number_of_letters**jnp.arange(2,motiflength+1),jnp.array([number_of_letters**(motiflength-1)]))))),dtype=int)
    ligation_spot_formations = jnp.zeros((jnp.sum(number_of_strands)+number_of_ends,jnp.sum(number_of_strands)+number_of_beginnings,jnp.sum(number_of_strands[1:])+number_of_beginnings+number_of_continuations+number_of_ends))

    for ligation_window_length in ligation_window_lengths:
        for ligation_spot in range(1,ligation_window_length-2):
            for g1,g2,g3,g4 in iterprod([0,1],repeat=4):
                if ligation_window_length != maximum_ligation_window_length:
                    if (g2,g3)==(1,1) or (g1,g4)==(1,1):
                        continue
                if g1 == 0:
                    # length-{ligation_spot}-strand
                    left_strandlength = ligation_spot
                    left_ligs_slices = [(slice(reactant_indices[left_strandlength-1],reactant_indices[left_strandlength]),)]
                    left_prods_slices = [
                            (0,) + (slice(None),)*int(left_strandlength)
                            ]
                    left_prods_zeros_shapes = [
                            (number_of_letters,)*int(left_strandlength)
                            ]
                    left_prods_zeros_reshapes = [
                            (number_of_letters**left_strandlength,)
                            ]
                    """
                    number_of_strands[ligation_spot]
                    left_indices = [slice(
                            jnp.sum(number_of_strands[:ligation_spot])-1,
                            jnp.sum(number_of_strands[:ligation_spot]) + number_of_strands[ligation_spot])
                            ]
                    """
                else:
                    # [strand[ii] for ii in range(ligation_spot+1,motiflength-1)] + ['end']
                    left_strandlengths = jnp.arange(ligation_spot+1,motiflength)
                    left_ligs_slices = [
                            (slice(reactant_indices[left_strandlength-1],reactant_indices[left_strandlength]),)
                            for  left_strandlength in left_strandlengths
                            ]
                    left_prods_slices = [
                        (None,)*int(left_strandlength-ligation_spot-1)
                        + (slice(1,None),)
                        + (slice(None),)*int(ligation_spot)
                        for left_strandlength in left_strandlengths
                    ]
                    left_prods_zeros_shapes = [
                            (number_of_letters,)*int(left_strandlength)
                            for left_strandlength in left_strandlengths
                            ]
                    left_prods_zeros_reshapes = [
                        (number_of_letters**left_strandlength,)
                        for left_strandlength in left_strandlengths
                    ]
                    # left_indices_collisions = [(number_of_letters,)*(motiflength-ligation_spot-2)+(slice(number_of_letters**(ligation_spot+1)),) for strandlength in range(ligation_spot,motiflength)]
                    """
                    left_indices_collisions_shapes = [(number_of_letters,)*strandlength for strandlength in range(ligation_spot, motiflength)]
                    left_indices_collisions_indices = [slice(
                        jnp.sum(number_of_letters**jnp.arange(1,strandlength)),
                        jnp.sum(number_of_letters**jnp.arange(1,strandlength+1)),
                        ) for strandlength in range(ligation_spot, motiflength)]
                    left_indices_productions = [(None,)*strandlength+(slice(number_of_letters**(ligation_spot+1)),) for strandlength in range(ligation_spot,motiflength)]#(motiflength-ligation_spot-2)
                    ligation_spot_formations[left_indices_collisions_indices,right_indices_collisions_indices,template_indices_collisions_indices].reshape(left_indices_collisions_shapes+right_indices_collisions_shapes,template_indices_collisions_shapes) = motif_production_rate_constants[left_indices_productions]
                    left_indices = [
                            slice(jnp.sum(number_of_strands[:ligation_spot])-1),None)
                            ]
                    collisions_format[left_indices_collisions] = motif_production_rate_constants[left_indices_productions]
                    """
                if g2 == 0:
                    # [strand[ligation_window_length-ls-2]]
                    right_strandlength = int(ligation_window_length-ligation_spot-2)
                    right_ligs_slices = [
                            (slice(reactant_indices[right_strandlength-1],reactant_indices[right_strandlength]),)
                            ]
                    right_prods_zeros_shapes = [(number_of_letters,)*int(right_strandlength)]
                    right_prods_zeros_reshapes = [(number_of_letters**right_strandlength,)]
                    right_prods_slices = [
                            (slice(None),)*right_strandlength+(0,)
                            ]
                else:
                    # right_categories = [strand[ii] for ii in range(ligation_window_length-ls-1,motiflength-1)] + ['beginning']
                    right_strandlengths = jnp.arange(ligation_window_length-ligation_spot-1,motiflength-1)
                    right_ligs_slices = [
                            (slice(reactant_indices[right_strandlength-1],reactant_indices[right_strandlength]),)
                            for right_strandlength in right_strandlengths
                            ]
                    right_prods_zeros_shapes = [(number_of_letters,)*int(right_strandlength) for right_strandlength in right_strandlengths]
                    right_prods_zeros_reshapes = [(number_of_letters**int(right_strandlength),) for right_strandlength in right_strandlengths]
                    right_prods_slices = [
                            (slice(None),)*int(ligation_window_length-ligation_spot-2)
                            + (slice(1,None),)
                            +(None,)*int(right_strandlength-(ligation_window_length-ligation_spot-2)-1)
                            for right_strandlength in right_strandlengths
                            ]
                if (g3,g4) == (0,0):
                    # template_category = strand[ligation_window_length-2]
                    template_strandlength = ligation_window_length-2
                    template_ligs_slices = [
                        (slice(template_indices[template_strandlength-1],template_indices[template_strandlength]),)
                    ]
                    template_prods_zeros_shapes = [(number_of_letters,)*(template_strandlength)]
                    template_prods_zeros_reshapes = [(number_of_letters**template_strandlength,)]
                    template_prods_slices = [
                            (0,)*(1-g3)
                            + (slice(None),)*int(ligation_window_length-2)
                            + (0,)*(1-g4)
                    ]
                elif (g3,g4) == (0,1):
                    # template_category = [strand[ii] for ii in range(ligation_window_length-1,motiflength-1)] + ['beginning']
                    template_strandlengths = jnp.arange(ligation_window_length-1,motiflength)
                    template_ligs_slices = [
                            (slice(template_indices[template_strandlength-1],template_indices[template_strandlength]),)
                            for template_strandlength in template_strandlengths
                            ]
                    template_prods_zeros_shapes = [(number_of_letters,)*int(template_strandlength) for template_strandlength in template_strandlengths]
                    template_prods_zeros_reshapes = [(number_of_letters**template_strandlength,) for template_strandlength in template_strandlengths]
                    template_prods_slices = [
                            (0,)*(1-g3)
                            + (slice(None),)*int(ligation_window_length-2)
                            + (0,)*(1-g4) + (slice(1,None),)*g4
                    ]
                elif (g3,g4) == (1,0):
                    # template_category = [strand[ii] for ii in range(ligation_window_length-1,motiflength-1)] + ['end']
                    template_strandlengths = jnp.arange(ligation_window_length-1,motiflength)
                    template_ligs_slices = [
                        (slice(template_indices[template_strandlength-1],template_indices[template_strandlength]),)
                        for template_strandlength in template_strandlengths
                        ]
                    template_ligs_slices[-1] = (slice(template_indices[-2],template_indices[-1]),)
                    template_prods_zeros_shapes = [(number_of_letters,)*int(template_strandlength) for template_strandlength in template_strandlengths]
                    template_prods_zeros_reshapes = [(number_of_letters**template_strandlength,) for template_strandlength in template_strandlengths]
                    template_prods_slices = [
                            (None,)*int(template_strandlength-ligation_window_length)
                            + (slice(1,None),)
                            + (slice(None),)*int(ligation_window_length-2)
                            + (0,)
                            for template_strandlength in template_strandlengths
                            ]
                elif (g3,g4) == (1,1):
                    # template_category = [strand[ii] for ii in range(ligation_window_length,motiflength-2)] + ['continuation'] + (lw<maximum_ligation_window_length)*['beginning','end']
                    template_strandlengths = jnp.concatenate((jnp.arange(ligation_window_length,motiflength+1),)+(jnp.array([motiflength-1]),)*(ligation_window_length<motiflength))
                    template_ligs_slices = [
                            (slice(template_indices[template_length_index-1],template_indices[template_length_index]),)
                            for template_length_index in range(ligation_window_length, ligation_window_length+len(template_strandlengths))
                            for left_continuation in jnp.arange(0,template_strandlengths[template_length_index]-ligation_window_length+1)
                            ]
                    template_prods_zeros_shapes = [
                            (number_of_letters,)*int(template_strandlengths[template_length_index])
                            for template_length_index in range(ligation_window_length, ligation_window_length+len(template_strandlengths))
                            for left_continuation in jnp.arange(0,template_strandlengths[template_length_index]-ligation_window_length+1)
                            ]
                    template_prods_zeros_reshapes = [
                            (number_of_letters**int(template_strandlengths[template_length_index]),)
                            for template_length_index in range(ligation_window_length, ligation_window_length+len(template_strandlengths))
                            for left_continuation in jnp.arange(0,template_strandlengths[template_length_index]-ligation_window_length+1)
                    ]
                    template_prods_slices = [
                        (None,)*int(left_continuation)
                        + (slice(1,None),)
                        + (slice(None),)*int(ligation_window_length-2)
                        + (slice(1,None),)
                        + (None,)*int(template_strandlength-ligation_window_length-left_continuation)
                        for template_length_index in range(ligation_window_length, ligation_window_length+len(template_strandlengths))
                        for left_continuation in jnp.arange(0,template_strandlengths[template_length_index]-ligation_window_length+1)
                    ]
                left_reactant_indices = jnp.arange(len(left_ligs_slices))
                right_reactant_indices = jnp.arange(len(right_ligs_slices))
                template_catalyst_indices = jnp.arange(len(template_ligs_slices))
                for left_reactant_ii, right_reactant_ii, template_ii in iterprod(left_reactant_indices, right_reactant_indices, template_catalyst_indices):
                    ligation_spot_formations = ligation_spot_formations.at[left_ligs_slices[left_reactant_ii]+right_ligs_slices[right_reactant_ii]+template_ligs_slices[template_ii]].add(
                        (jnp.zeros(left_prods_zeros_shapes[left_reactant_ii] + right_prods_zeros_shapes[right_reactant_ii] + template_prods_zeros_shapes[template_ii])
                         + motif_production_rate_constants[left_prods_slices[left_reactant_ii]+right_prods_slices[right_reactant_ii]+template_prods_slices[template_ii]]).reshape(
                             left_prods_zeros_reshapes[left_reactant_ii] + right_prods_zeros_reshapes[right_reactant_ii] + template_prods_zeros_reshapes[template_ii]
                             )
                         )
    return ligation_spot_formations

def motif_production_transition_kernel_from_motif_production_rate_constants_array(
        motif_production_rate_constants : jax.Array,
        number_of_letters : int = 4,
        motiflength : int = 4,
        maximum_ligation_window_length : int = 4
        ) -> jax.Array:
    raise NotImplementedError("motif production transition kernel under construction")
    extended_motif_production_rate_constants = _extend_motif_production_rate_constants_array_to_collisions_format(motif_production_rate_constants, 
        maximum_ligation_window_length = maximum_ligation_window_length,
        motiflength = motiflength,
        number_of_letters = number_of_letters,)
    transition_matrix = motif_production_transition_kernel_matrix(
            number_of_letters,
            motiflength,
            maximum_ligation_window_length
            )
    return transition_matrix.reshape((transition_matrix.shape[0],-1)) @ extended_motif_production_rate_constants.flatten()

def motif_production_transition_kernel_matrix(
        number_of_letters : int = 4,
        motiflength : int = 4,
        maximum_ligation_window_length : int = 4
        ) -> jax.Array:
    """
    Motif productions are in the
    Third order Motif Collisions include all combinations of three motifs independent of their category.
    Ligation sites are a subclass third order motif collisions,
    where the first motif (the left or fore motif) is a strand or an ending,
    the second motif (right or rear motif) is a strand or a beginning
    and the third motif (the template) is a strand of at least length 2 or beginning, continuation or beginning.

    Motif Productions sights only need to consider the hybridization sight and eventual dangling ends.
    They are thus a subclass of ligation sites.
    Ligation Windows categorize the motif production sites.
    The ligation windows have the length of the hybridization sight plus 2 more nucleotides to cover dangling ends.
    The 'maximum_ligation_window_length' is maximally equal to the motiflength,
    as longer motifs, and thus longer hybridization sights are not tracked.
    The biggest ligation window has thus the 'maximum_ligation_window_length'.
    The biggest ligation window is the only one that can contain hybridization sights of length equal to the ligation window or one nucleotide less.
    In addition to the hybridization site, each ligation window also stores a number between 1 and the length of the hybridization site minus one that indicates the 'ligation_spot'.

    The motif_production_rate_constants are categorized by the ligation windows.
    For the transition kernal they first need to be extended to the shape of the ligation_sites,
    which means a degeneracy, i.e. all ligation_sites with the same ligation window get the same contribution.

    To set up the transition kernel,
    all reactants in the corresponding ligation_site get a negative contribution of the corresponding rate,
    all products get a positive contribution.

    Parameters:
    -----------
    motif_production_rate_constants : jax.Array
        The rate constants for the motif production
    number_of_letters : int
        The number of letters in the alphabet
    motiflength : int
        The length of the motif
    maximum_ligation_window_length : int
        Maximum ligaion window length, default is 4

    Returns:
    --------
    jax.Array
        The motif production transition kernel
    """
    raise NotImplementedError("infer.motif_production_transition_kernel_matrix currently under construction")
    number_of_strands = int(jnp.sum(number_of_letters**jnp.arange(1,motiflength-1)))
    number_of_endings = number_of_beginnings = int(number_of_letters**(motiflength-1))
    number_of_continuations = int(number_of_letters**motiflength)
    number_of_reactants = int(jnp.sum(number_of_letters**jnp.arange(1,motiflength)))
    number_of_templates = int(jnp.sum(number_of_letters**jnp.arange(1,motiflength+1))+number_of_letters**(motiflength-1)-number_of_letters)

    extended_motif_production_rate_constants_shape = (number_of_reactants,)*2+(number_of_templates,)
    motif_production_transition_kernel = jnp.zeros((jnp.sum(motiflength**jnp.arange(1,motiflength+1))+number_of_letters**(motiflength-1),)+extended_motif_production_rate_constants_shape)

    left_reactant_entries = -jnp.outer(jnp.fill_diagonal(jnp.zeros((number_of_reactants,)*2),1,inplace=False),jnp.ones((number_of_reactants,number_of_templates))).reshape((jnp.sum(number_of_letters**jnp.arange(1,motiflength)),)*2+(number_of_reactants,number_of_templates))
    right_reactant_entries = jnp.moveaxis(left_reactant_entries,source=1,destination=2)

    left_reactant_slices = [
            (slice(number_of_strands,),slice(number_of_strands,),slice(None),slice(None)),
            (slice(number_of_strands+number_of_beginnings+number_of_continuations,None),slice(number_of_strands,number_of_reactants),slice(None),slice(None))
            ]
    left_reactants_entries_slices = [(slice(number_of_strands),)*2, (slice(number_of_strands,number_of_reactants),)*2]
    for ii in range(len(left_reactant_slices)):
        motif_production_transition_kernel = motif_production_transition_kernel.at[left_reactant_slices[ii]].add(
                left_reactant_entries[left_reactants_entries_slices[ii]]
                )

    right_reactant_slices = (slice(number_of_reactants,),slice(None),slice(number_of_reactants,),slice(None))
    motif_production_transition_kernel = motif_production_transition_kernel.at[right_reactant_slices].add(
            right_reactant_entries
            )

    motif_indices = jnp.concatenate([jnp.zeros(1),jnp.cumsum(number_of_letters**jnp.arange(1,motiflength+1)),jnp.array([jnp.sum(number_of_letters**jnp.arange(1,motiflength+1))+number_of_letters**(motiflength-1)])],)
    for strandlength in range(2,motiflength+1):
        if strandlength == motiflength-1:
            for ligation_spot in range(1,motiflength-2):
                # case1: beginning product
                for left_reactant_length in range(ligation_spot,motiflength-1):
                    for right_reactant_length in range(strandlength-left_reactant_length,motiflength):
                        product_slices = (
                                slice(int(motif_indices[-3-1]),int(motif_indices[-3])),
                                slice(int(motif_indices[left_reactant_length-1]),int(motif_indices[left_reactant_length])),
                                slice(int(motif_indices[right_reactant_length-1]),int(motif_indices[right_reactant_length])),
                                slice(None)
                                )
                        product_entries = jnp.outer(
                                jnp.fill_diagonal(jnp.zeros((number_of_letters**strandlength,)*2),1,inplace=False),
                                jnp.ones((int(number_of_letters**(right_reactant_length-(strandlength-left_reactant_length))),number_of_templates,))
                                ).reshape(
                                        (number_of_letters**strandlength,number_of_letters**left_reactant_length,number_of_letters**right_reactant_length,number_of_templates,)
                                        )
                        motif_production_transition_kernel = motif_production_transition_kernel.at[product_slices].add(
                                product_entries 
                                )
                # case2: ending product
                for left_reactant_length in range(ligation_spot,motiflength):
                    for right_reactant_length in range(max(strandlength-left_reactant_length,1), strandlength):
                        product_slices = (
                                slice(int(motif_indices[-1-1]),int(motif_indices[-1])),
                                slice(int(motif_indices[left_reactant_length-1]),int(motif_indices[left_reactant_length])),
                                slice(int(motif_indices[right_reactant_length-1]),int(motif_indices[right_reactant_length])),
                                slice(None)
                                )
                        zeros_shape = (number_of_letters**strandlength,)*2
                        ones_shape = (number_of_templates,int(number_of_letters**max(left_reactant_length+right_reactant_length-strandlength,0))) #Fixme: max(left_reactant_length-ligation_spot-k,0)???, k=1? oder k=l-3 oder so?
                        print(f'{strandlength=}')
                        print(f'{left_reactant_length=}, {right_reactant_length=}')
                        print(f'{zeros_shape=}')
                        print(f'{ones_shape=}')
                        print(f'{product_slices=}')
                        product_entries = jnp.moveaxis(
                                jnp.outer(
                                    jnp.fill_diagonal(jnp.zeros(zeros_shape),1,inplace=False),
                                    jnp.ones(ones_shape)
                                    ).reshape(zeros_shape+ones_shape),
                                source=-1,
                                destination=1,
                                ).reshape(
                                        (number_of_letters**strandlength,number_of_letters**left_reactant_length,number_of_letters**right_reactant_length,number_of_templates,)
                                        )
                        motif_production_transition_kernel = motif_production_transition_kernel.at[product_slices].add(
                                product_entries
                                )
        elif strandlength == motiflength:
            for ligation_spot in range(1,motiflength-2):
                for left_reactant_length in range(ligation_spot,motiflength):
                    for right_reactant_length in range(strandlength-left_reactant_length,motiflength):
                        product_slices = (
                                slice(int(motif_indices[-2-1]),int(motif_indices[-2])),
                                slice(int(motif_indices[left_reactant_length-1]),int(motif_indices[left_reactant_length])),
                                slice(int(motif_indices[right_reactant_length-1]),int(motif_indices[right_reactant_length])),
                                slice(None)
                                )
                        for product_index in range(0,left_reactant_length+right_reactant_length-motiflength+1):
                            # product entries have to be for all products from left to right
                            source_axes = jnp.arange(1, 1+product_index)[::-1]
                            destination_axes = jnp.arange(0,product_index)
                            zeros_shape = (number_of_letters**strandlength,)*2
                            ones_shape = (int(number_of_letters**(left_reactant_length+right_reactant_length-strandlength-product_index)), number_of_templates, int(number_of_letters**product_index))
                            product_entries = jnp.moveaxis(
                                    jnp.outer(
                                        jnp.fill_diagonal(jnp.zeros(zeros_shape),1,inplace=False),
                                        jnp.ones(ones_shape)
                                        ).reshape(zeros_shape+ones_shape),
                                    source=-1,
                                    destination=1,
                                    ).reshape(
                                            (number_of_letters**strandlength,number_of_letters**left_reactant_length,number_of_letters**right_reactant_length,number_of_templates,)
                                            )
                            motif_production_transition_kernel = motif_production_transition_kernel.at[product_slices].add(
                                    product_entries 
                                    )
        else:
            for ligation_spot in range(1,strandlength):
                left_reactant_length = ligation_spot
                right_reactant_length = strandlength - ligation_spot
                product_slices = (
                    slice(int(motif_indices[strandlength-1]),int(motif_indices[strandlength])),
                    slice(int(motif_indices[left_reactant_length-1]),int(motif_indices[left_reactant_length])),
                    slice(int(motif_indices[right_reactant_length-1]),int(motif_indices[right_reactant_length])),
                    slice(None))
                product_entries = jnp.outer(jnp.fill_diagonal(jnp.zeros((number_of_letters**strandlength,)*2),1,inplace=False),jnp.ones((number_of_templates,))).reshape((number_of_letters**strandlength,number_of_letters**ligation_spot,number_of_letters**(strandlength-ligation_spot),number_of_templates,))
                motif_production_transition_kernel = motif_production_transition_kernel.at[product_slices].add(
                        product_entries 
                        )
    return sparse.BCOO.fromdense(motif_production_transition_kernel)



import numpy as np
import nifty8 as ift
from collections import namedtuple
from ..obj.units import make_unit, Unit

from ..domains.motif_space import _return_motif_categories
from ..domains.motif_breakage_space import MotifBreakageSpace

def MotifBreakageVector(motiflength : int,
                        alphabet : list,
                        unit : Unit):
    motif_categories = _return_motif_categories(motiflength=motiflength)[1:]
    motif_vector_properties = {'motiflength' : motiflength,
        'alphabet' : alphabet,
        'number_of_letters' : len(alphabet),
        'unit' : unit}
    def makeMotifBreakageVector(motif_breakage_vector_dct : dict):
        breakages = ift.MultiField.from_raw(MotifBreakageSpace.make(alphabet, motiflength),
                                            motif_breakage_vector_dct)
        motif_vector = namedtuple('MotifBreakageVector',
                ('breakages',) + tuple(motif_vector_properties.keys()))
        return motif_vector(**{**{'breakages': breakages}, **motif_vector_properties})
    return makeMotifBreakageVector

def _create_empty_motif_breakage_dct(motiflength : int,
        alphabet : list = ['a','b']) -> dict:
    number_of_letters = len(alphabet)

    empty_motif_vector = {}
    for strandlength in range(2,motiflength-1):
        category = _return_motif_categories()[0].format(strandlength)
        shape = (number_of_letters,)*strandlength
        for hyphen_index in range(1,strandlength):
            empty_motif_vector[category+'_{}'.format(hyphen_index)] = np.zeros(shape)
    motif_categories = _return_motif_categories(motiflength=motiflength)
    for category in motif_categories[-3:]:
        category_length = motiflength-1 + (category==motif_categories[-2])
        shape = (number_of_letters,)*category_length
        for hyphen_index in range(category_length):
            empty_motif_vector[category + '_{}'.format(hyphen_index)] = np.zeros(shape)
    return empty_motif_vector

def _breakage_array_indices(
        motiflength : int,
        strandlength : int,
        breakage_spot : int,
        sequence_is_beginning : bool
    ) -> tuple:
    """
    sets up the indices of the breakage array given the parameters.
    """
    end_overlap = max(0,strandlength-breakage_spot-motiflength//2)
    beginning_overlap = max(0, breakage_spot-motiflength+motiflength//2)
    end_length_without_overlap = strandlength-breakage_spot-end_overlap
    beginning_length_without_overlap = min(breakage_spot, motiflength-motiflength//2)
    zeros_between_end_overlap_and_beginning = max(int(sequence_is_beginning),motiflength-motiflength//2-end_overlap-beginning_length_without_overlap)
    zeros_between_end_and_beginning_overlap = max(0,motiflength//2-end_length_without_overlap-beginning_overlap)

    indices_sequence_arrays =  (slice(1,None),)*end_overlap
    indices_sequence_arrays +=  (0,)*zeros_between_end_overlap_and_beginning
    indices_sequence_arrays += (slice(1,None),)*(beginning_length_without_overlap-1)
    indices_sequence_arrays += (slice(None),)*2
    indices_sequence_arrays += (slice(1,None),)*(end_length_without_overlap-1)
    indices_sequence_arrays += (0,)*zeros_between_end_and_beginning_overlap
    indices_sequence_arrays += (slice(1,None),)*beginning_overlap
    return indices_sequence_arrays

def _transform_sequence_array_to_motif_breakage_array(
        sequence_array,
        motiflength : int,
        breakage_spot : int,
        sequence_is_beginning : bool = True
        ) -> np.ndarray:
    """
    transforms an array that only tracks letters (occupied nucleotides)
    into an array that explicitely tracks empty spots with zeros.
    Note that the second spot is always a letter, though.

    Parameters:
    -----------
    sequence_array : nd-array
    motiflength : int
    sequence_is_beginning : boolean (optional)
        whether the sequence is beginning, i.e. the first spot is zero,
        else it will be considered as end or continuation if all spots are
        letters
        default : True

    Returns:
    --------
    motif_array : nd-array
    """
    if len(sequence_array.shape)==motiflength:
        sequence_is_beginning = False
    number_of_letters = sequence_array.shape[0]
    strandlength = len(sequence_array.shape)

    indices_sequence_arrays = _breakage_array_indices(motiflength, strandlength, breakage_spot, sequence_is_beginning)

    motif_array_shape = (number_of_letters+1,)*int(motiflength>2)
    motif_array_shape += (number_of_letters,)*2
    motif_array_shape += (number_of_letters+1,)*(motiflength-2-int(motiflength>2))
    motif_array = np.zeros(motif_array_shape)

    motif_array[indices_sequence_arrays] = sequence_array
    return motif_array

def _motif_breakage_vector_as_array(
        motif_breakage_vector : MotifBreakageVector
        ) -> np.ndarray:
    """
    transforms a motif vector into a numpy-array

    Parameters:
    -----------
    motif_breakage_vector : MotifBreakageVector

    Returns:
    --------
    motif_breakage_array : np.ndarray
    """
    motiflength = motif_breakage_vector.motiflength
    number_of_letters = motif_breakage_vector.number_of_letters

    beginning_breakage_spots = (1, motiflength-motiflength//2)
    continuation_breakage_spots = (motiflength-motiflength//2,motiflength-motiflength//2+1)
    end_breakage_spots = (motiflength-motiflength//2, motiflength-1)
    bs = [beginning_breakage_spots, continuation_breakage_spots, end_breakage_spots]
    motif_categories = [(_return_motif_categories()[ii], breakage_spot) for ii in range(1,4) for breakage_spot in range(*bs[ii-1])]

    strand_categories = [(_return_motif_categories()[0].format(strandlength), breakage_spot) for strandlength in range(1,motiflength-1) for breakage_spot in range(1,strandlength)]

    motif_array_shape = (number_of_letters+1,)*int(motiflength>2)
    motif_array_shape += (number_of_letters,)*2
    motif_array_shape += (number_of_letters+1,)*(motiflength-2-int(motiflength>2))
    motif_array = np.zeros(motif_array_shape)

    for motif_category in strand_categories:
        breakage_spot = motif_category[1]
        motif_category = motif_category[0]+'_{}'.format(breakage_spot)
        motif_array += _transform_sequence_array_to_motif_breakage_array(
            motif_breakage_vector.breakages[motif_category].val,
            motiflength,
            breakage_spot,
            sequence_is_beginning = True
        )
    for motif_category in motif_categories:
        breakage_spot = motif_category[1]
        sequence_is_beginning = motif_category[0]=='beginning'
        motif_category = motif_category[0]+'_{}'.format(breakage_spot)
        motif_array += _transform_sequence_array_to_motif_breakage_array(
            motif_breakage_vector.breakages[motif_category].val,
            motiflength,
            breakage_spot,
            sequence_is_beginning = sequence_is_beginning
        )
    return motif_array

def _array_to_motif_breakage_vector(
        motif_breakage_vector_array : np.ndarray,
        motiflength : int,
        alphabet : list,
        unit : Unit
        ) -> MotifBreakageVector:
    if motiflength != len(motif_breakage_vector_array.shape):
        raise ValueError("motiflength inconsistent with array shape")
    number_of_letters = len(alphabet)
    if number_of_letters != motif_breakage_vector_array.shape[0]-1:
        raise ValueError("alphabet shape ({alphabet_shape}) inconsistent with motif_breakage_vector_array shape ({motif_vector_array_shape})".format(
            alphabet_shape = len(alphabet),
            motif_vector_array_shape = motif_breakage_vector_array.shape
                    ))

    beginning_breakage_spots = (1, motiflength-motiflength//2)
    continuation_breakage_spots = (motiflength-motiflength//2,motiflength-motiflength//2+1)
    end_breakage_spots = (motiflength-motiflength//2, motiflength-1)
    bs = [beginning_breakage_spots, continuation_breakage_spots, end_breakage_spots]

    motif_categories = [(_return_motif_categories()[ii], breakage_spot) for ii in range(1,4) for breakage_spot in range(*bs[ii-1])]
    strand_categories = [(_return_motif_categories()[0].format(strandlength), strandlength, breakage_spot) for strandlength in range(1,motiflength-1) for breakage_spot in range(1,strandlength)]

    motif_vector_dct = {}
    for motif_category in strand_categories:
        breakage_spot = motif_category[2]
        strandlength = motif_category[1]
        motif_category = motif_category[0]+'_{}'.format(breakage_spot)
        current_indices = _breakage_array_indices(motiflength, strandlength, breakage_spot, sequence_is_beginning=True)
        motif_vector_dct[motif_category] = motif_breakage_vector_array[current_indices]
    for motif_category in motif_categories:
        breakage_spot = motif_category[1]
        sequence_is_beginning = motif_category[0]=='beginning'
        strandlength = motiflength-int(motif_category[0]!='continuation')
        motif_category = motif_category[0]+'_{}'.format(breakage_spot)
        current_indices = _breakage_array_indices(motiflength, strandlength, breakage_spot, sequence_is_beginning)
        motif_vector_dct[motif_category] = motif_breakage_vector_array[current_indices]
    return motif_vector_dct

def isinstance_motifbreakagevector(obj) -> bool:
    is_motif_breakage_vector = True
    keys = ['motiflength','alphabet','number_of_letters','unit','breakages']
    for key in obj._asdict().keys():
        if key not in keys:
            print('Not a MotifBreakageVector, missing key: {}.'.format(key))
            return False
    keys = MotifBreakageSpace.make(obj.alphabet, obj.motiflength).keys()
    for key in obj.breakages.keys():
        if key not in keys:
            print('Not a MotifBreakageVector, missing key in breakages field: {}.'.format(key))
            return False
    is_motif_breakage_vector *= isinstance(obj, tuple)
    return bool(is_motif_breakage_vector)



import numpy as np

from ..obj.motif_trajectory import MotifTrajectory, _motif_trajectory_as_array

def system_level_motif_zebraness(
        motif_trajectory : MotifTrajectory,
        axis=0
    ) -> np.ndarray:
    c, _ = _motif_trajectory_as_array(motif_trajectory)
    shape = c.shape
    if axis == 0:
        c = c.reshape((1,)+shape)
    if axis > 1:
        raise NotImplementedError("compute_system_level_motif_zebraness so for only defined for maximum first axis being time axis or static motif concentration vector")
    shape = c.shape
    if shape[1]>3:
        raise NotImplementedError("Zebraness only defined for binary alphabet.")
    if len(shape)!=5:
        raise NotImplementedError("Zebraness only implemented for four-letter motifs.")
    number_of_timesteps = shape[0]

    concentration_of_zebra_ligations = np.zeros(number_of_timesteps)
    concentration_of_zebra_ligations += np.sum(c[:,:,-2,-1,:].reshape((number_of_timesteps,-1)), axis =1)
    concentration_of_zebra_ligations += np.sum(c[:,:,-1,-2,:].reshape((number_of_timesteps,-1)), axis=1)

    concentration_of_ligations_in_total = np.zeros(number_of_timesteps)
    concentration_of_ligations_in_total += np.sum(
                c[:,:,-2:,-2:,:].reshape((number_of_timesteps,-1)),
                axis=1
                )

    return concentration_of_zebra_ligations/concentration_of_ligations_in_total

def individual_motif_zebraness(c):
    shape = c.shape
    if len(shape)==4:
        c = c.reshape((-1,)+shape)
    if shape[1]>3:
        raise NotImplementedError("Zebraness only defined for binary alphabet.")
    if len(shape)!=5:
        raise NotImplementedError("Zebraness only implemented for four-letter motifs.")
    # Monomers don't have zebraness
    z = np.zeros(shape)
    # Zebraness of first binary motif
    z[:,-1,-2,-2:,:]+=1
    z[:,-2,-1,-2:,:]+=1
    # Zebraness of second binary motif
    z[:,:,-1,-2,:]+=1
    z[:,:,-2,-1,:]+=1
    # Zebraness of last binary motif
    z[:,:,-2:,-1,-2]+=1
    z[:,:,-2:,-2,-1]+=1
    # multiply by length
    y = np.zeros(shape)
    y[:,0,-2:,-2:,0] = 1.
    y[:,0,-2:,-2:,-2:] = 2.
    y[:,-2:,-2:,-2:,0] = 2.
    y[:,-2:,-2:,-2:,-2:] = 3.
    return z/y



from .logc import unzero

from .manage_strand_reactor_files import _return_parameters_to_read_from_parameters_file
from .manage_strand_reactor_files import _create_typical_strand_reactor_dirpath
from .manage_strand_reactor_files import _create_typical_strand_trajectory_ensemble_dirpath
from .manage_strand_reactor_files import _create_typical_strand_trajectory_dirpath
from .manage_strand_reactor_files import _create_typical_strand_trajectory_section_dirpath
from .manage_strand_reactor_files import _create_typical_strand_trajectory_section_complexes_filepath
from .manage_strand_reactor_files import _create_typical_strand_parameters_filepath
from .manage_strand_reactor_files import _create_typical_alphabet_filepath
from .manage_strand_reactor_files import _create_complexes_filepath_lists
from .manage_strand_reactor_files import _create_ligations_filepath_lists
from .manage_strand_reactor_files import transform_execution_time_txt_to_list

from .save import create_strand_trajectory_ensemble_path
from .save import create_trajectory_ensemble_path
from .save import create_motif_production_trajectory_ensemble_path

from .split_config import split_config



from ..read.strand_reactor_parameters import strand_reactor_parameters as read_strand_reactor_parameters
from ..utils.manage_strand_reactor_files import _create_typical_strand_parameters_filepath
from ..infer.motif_production_rate_constants_from_strand_reactor_parameters import motif_production_rate_constants_from_strand_reactor_parameters as infer_motif_production_rate_constants_from_strand_reactor_parameters
from .alphabet import alphabet as get_alphabet

from ..obj.motif_production_vector import MotifProductionVector


def motif_production_rate_constants_from_strand_reactor_parameters(
        strand_trajectory_id,
        motiflength : int,
        complements : list,
        param_file_no : int = 0,
        maximum_ligation_window_length : int = None
    ) -> MotifProductionVector:
    if motiflength != 4:
        raise NotImplementedError("")
    if maximum_ligation_window_length is None:
        maximum_ligation_window_length = motiflength
    filepath = _create_typical_strand_parameters_filepath(strand_trajectory_id, param_file_no = param_file_no)
    strand_reactor_parameters = read_strand_reactor_parameters(filepath)
    alphabet = get_alphabet(strand_trajectory_id)
    return infer_motif_production_rate_constants_from_strand_reactor_parameters(
        strand_reactor_parameters,
        motiflength,
        alphabet,
        maximum_ligation_window_length,
        complements
    )



import jax.numpy as jnp
from jax.experimental import sparse
from itertools import product as iterprod

from .motif_production_transition_kernel import _hybridization_site_categories


def motif_production_rate_constants_extension_matrix(
        number_of_letters : int = 2,
        motiflength : int = 4,
        hybridization_length_max : int = 4
        ) -> sparse.BCOO:
    """
    generates a matrix as jax.experimental.sparse.BCOO matrix
    that transforms hybridization site formated vectors into collision format.

    Parameters:
    -----------
    number_of_letters : int
        default : 2
    motiflength : int
        default : 4
    hybridization_length_max : int
        default : 4
    """
    raise NotImplementedError("motif_production_rate_constants_extension_matrix is under construction")
    hybridization_site_caterories, hybridization_configuration_indices = _hybridization_site_categories(number_of_letters, hybridization_length_max)
    motif_indices = jnp.concatenate([jnp.zeros(1),jnp.cumsum(number_of_letters**jnp.arange(1,motiflength+1)),jnp.array([jnp.sum(number_of_letters**jnp.arange(1,motiflength+1))+number_of_letters**(motiflength-1)])],dtype=int)
    template_indices = (motif_indices-motif_indices[1]).at[0].set(0)
    number_of_left_reactants = number_of_right_reactants = int(motif_indices[-3])
    number_of_templates = int(template_indices[-1])
    mprcem_shape = (number_of_left_reactants,number_of_right_reactants, number_of_templates, hybridization_configuration_indices[-1])
    motif_production_rate_constants_extension_matrx = sparse.BCOO((jnp.array([0]),jnp.array([[0]*len(mprcem_shape)])),shape=mprcem_shape)
    for a_index in range(len(hybridization_site_caterories)):
        a = hybridization_site_caterories[a_index]
        left_shift = a[-1]
        left_ligant_length = a[1]
        right_ligant_length = a[2]
        template_length = a[3]
        right_shift = left_shift + right_ligant_length + left_ligant_length - template_length
        if left_shift==-1:
            left_ligant_lengths = jnp.arange(a[1],motiflength)
            # blunt end gets extended iff hybridization_length=hybridization_length_max and the ligation spot is at the center or further apart.
            # for the other part of the complex, the same rules apply as for every hybridization_length: for a dangling end, the dangling strand/motif can be extended
            # for a blunt ent that is close to the ligation spot, the complex is supposed to end there
            # This way, we only extend blunt ends on both sides, if the ligation_spot is at the center
            # If hybridization_length_max is uneven, the right central ligation spot is treated as the center,
            # so the center is at hybridization_length_max-hybridization_length_max//2 from the left side and hybridization_lengh_max//2 from the right side apart 
        elif (a[0]==hybridization_length_max) and (left_shift==0) and (a[1]>=hybridization_length_max-hybridization_length_max//2):
            # continue left ligant
            left_ligant_lengths = jnp.arange(a[1],motiflength)
        else:
            left_ligant_lengths = [a[1]]
        if right_shift==1:
            right_ligant_lengths = jnp.arange(a[2],motiflength)
        elif (a[0]==hybridization_length_max) and (right_shift==0) and (a[2]>=hybridization_length_max//2):
            right_ligant_lengths = jnp.arange(a[2],motiflength)
        else:
            right_ligant_lengths = [a[2]]
        template_might_continue_forwards = (left_shift==1) or ((a[0]==hybridization_length_max) and (left_shift==0))
        template_might_continue_backwards = (right_shift==-1) or ((a[0]==hybridization_length_max) and (right_shift==0))
        template_might_continue_both_ways = (template_might_continue_forwards and template_might_continue_backwards)
        if template_might_continue_both_ways:
            template_overhangs = jnp.arange(motiflength-a[3])
        elif template_might_continue_forwards: #(but not backwards)
            template_overhangs = jnp.arange(motiflength-a[3])
        else:
            template_overhangs = jnp.arange(1)
        for left_ligant_length, right_ligant_length, template_overhang in iterprod(left_ligant_lengths, right_ligant_lengths, template_overhangs):
            if template_overhang+a[3] == motiflength:
                template_lengths = jnp.array([motiflength])
            elif template_might_continue_backwards or template_might_continue_forwards:
                template_lengths = jnp.arange(a[3]+template_overhang, motiflength)
            else:
                template_lengths = jnp.array([a[3]+template_overhang])
            for template_length in template_lengths:
                if (template_length == (motiflength-1)) and template_might_continue_backwards and not template_might_continue_forwards:
                    template_index_start = template_indices[-1-1]
                else:
                    template_index_start = template_indices[template_length-1]

                matrix_indices = [
                        (il*number_of_letters**(a[1])+i1+motif_indices[left_ligant_length-1],
                         i2*number_of_letters**(right_ligant_length-a[2])+ir+motif_indices[right_ligant_length-1],
                         int(itc*number_of_letters**(template_overhang+a[3])+i3*number_of_letters**(template_overhang)+io+template_index_start),
                         int(i1*number_of_letters**(a[2]+a[3])+i2*number_of_letters**a[3]+i3+hybridization_configuration_indices[a_index]))
                        for i1,i2,i3 in iterprod(
                            range(number_of_letters**a[1]),
                            range(number_of_letters**a[2]),
                            range(number_of_letters**a[3])
                            )
                        for il,ir,io,itc in iterprod(
                            range(number_of_letters**int(left_ligant_length-a[1])),
                            range(number_of_letters**int(right_ligant_length-a[2])),
                            range(number_of_letters**int(template_overhang)),
                            range(number_of_letters**int(template_length-template_overhang-a[3])))
                        ]
                motif_production_rate_constants_extension_matrx = motif_production_rate_constants_extension_matrx + sparse.BCOO((jnp.array([1]*len(matrix_indices)),matrix_indices),shape=(mprcem_shape))
    return motif_production_rate_constants_extension_matrx



import nifty8 as ift
from typing import Union

from .time_space import TimeSpace, TimeRGSpace, UnstructuredTimeDomain

def TrajectorySpace(motif_space : Union[ift.MultiDomain, ift.DomainTuple, ift.Domain],
                    time_space : Union[TimeRGSpace,UnstructuredTimeDomain]) -> ift.MultiField:
    while isinstance(motif_space, ift.DomainTuple):
        motif_space = motif_space[0]
    while isinstance(time_space, ift.DomainTuple):
        time_space = time_space[0]
    trajectory_space = {}
    for key in motif_space.keys():
        ms = motif_space[key]
        if isinstance(ms, ift.DomainTuple):
            ms = tuple([mm for mm in ms])
        else:
            ms = (ms,)
        trajectory_space[key] = ift.DomainTuple.make((time_space,)+ms)
    return ift.MultiDomain.make(trajectory_space)
